//********************************************************************
UNIT GUI;
//********************************************************************
INTERFACE

USES
   SysUtils, Types, Classes, Contnrs,
   dglDataProvider, OSSupport, GraphicSupport;

//********************************************************************
TYPE
   // Widget-alignment
   TAlignment = (alNone,alAll,alLeft,alTop,alRight,alBottom);
   // Text-alignment
   TTextAlignment = (taTopLeft,taTopRight,taTopCenter,
                     taMidLeft,taMidRight,taMidCenter,
                     taBottomLeft,taBottomRight,taBottomCenter);

   TWindowRect = TRect;

   TNotifyEvent = Procedure Of Object;
   TKeyEvent = Procedure(AKey,PreviousKey: LongWord) Of Object;
   TMouseBtnEvent = Procedure(X,Y,AddKeys: Integer) Of Object;
   TMouseMoveEvent = Procedure(X,Y: Integer) Of Object;
   TMouseWheelEvent = Procedure(X,Y,ADelta: Integer) Of Object;
   TResizeEvent = Procedure(X,Y: Integer) Of Object;

//********************************************************************
   TThemeDef = Record
      thClass: String;
      thColor: String;
      thFont: String;
      thImage: String;
      thShape: String;
   End;
   PThemeDef = ^TThemeDef;
//********************************************************************
   TTheme = Class(TPersistent)
   Private
      fThemeDef: TList;

      Procedure AddThemeClass(AThemeClass: String);
      Function GetThemeClass(AIndex: Integer): String;
      Function GetThemeCount: Integer;

      Function GetThemeColor(AThemeClass: String): String;
      Function GetThemeFont(AThemeClass: String): String;
      Function GetThemeImage(AThemeClass: String): String;
      Function GetThemeShape(AThemeClass: String): String;

      Function GetThemeColorByIndex(AIndex: Integer): String;
      Function GetThemeFontByIndex(AIndex: Integer): String;
      Function GetThemeImageByIndex(AIndex: Integer): String;
      Function GetThemeShapeByIndex(AIndex: Integer): String;

      Function ThemeDefByIndex(AIndex: Integer): PThemeDef;
      Function ThemeDefByName(AThemeClass: String): PThemeDef;
   Public
      Property Count: Integer Read GetThemeCount;
      Property ColorByClassName[Index: String]: String Read GetThemeColor;
      Property FontByClassName[Index: String]: String Read GetThemeFont;
      Property ImageByClassName[Index: String]: String Read GetThemeImage;
      Property ShapeByClassName[Index: String]: String Read GetThemeShape;

      Property ColorByIndex[Index: Integer]: String Read GetThemeColorByIndex;
      Property FontByIndex[Index: Integer]: String Read GetThemeFontByIndex;
      Property ImageByIndex[Index: Integer]: String Read GetThemeImageByIndex;
      Property ShapeByIndex[Index: Integer]: String Read GetThemeShapeByIndex;

      Constructor Create;
      Destructor Destroy; Override;
      Procedure Load;
      Procedure Save;
   End;

//********************************************************************
   TGUIItem = Class(TIDObject)
   Private
      // Private user's event-methods
      fOnKeyDown: TKeyEvent;
      fOnKeyUp: TKeyEvent;
      fOnMouseBtnLDown: TMouseBtnEvent;
      fOnMouseBtnLUp: TMouseBtnEvent;
      fOnMouseBtnRDown: TMouseBtnEvent;
      fOnMouseBtnMDown: TMouseBtnEvent;
      fOnMouseMove: TMouseMoveEvent;
      fOnMouseWheelDown: TMouseWheelEvent;
      fOnMouseWheelUp: TMouseWheelEvent;

      // Private GUI-event-methods
      fOnIdle: TNotifyEvent;
      fOnPaint: TNotifyEvent;
      fOnResize: TResizeEvent;

      // Visual variables
      fVisible: Boolean;

      fLeft,fTop,fWidth,fHeight: Integer;

      fBGColor: TRGBAColor;   // BackGroundColor
      fFGColor: TRGBAColor;   // ForeGroundColor
      fFont: TFreeType2Font;
      fImage: TRGBAImage;
      fShape: TShape;

      fGraphTool: TGraphManager;
      fTheme: TTheme;

   Protected
      fRoot: TGUIItem;
      fParent: TGUIItem;
      fChildren: TObjectList;
      fFocusedItem: TGUIItem;
      fHoveredItem: TGUIItem;
      fRepeat: Boolean;

      // Visual variables
      fDragMode,fStartX,fStartY: Word;
      fDrawMode: TDrawMode;
      fFrameState: TFrameState;
      fFrameWidth: Integer;
      fFadeIn: Boolean;
      fAlign: TAlignment;
      fCaption: String;
      fTextAlign: TTextAlignment;
      fTextPos: TPoint;

       Function Hit(X,Y: Integer): TGUIItem; Virtual;
      Procedure Move(X,Y: Integer); Virtual;
      Procedure Paint; Virtual;
      Procedure Resize(X,Y: Integer); Virtual;

       Function TREEHit(X,Y: Integer): TGUIItem;
       Function TREEItemByName(AName: String): TGUIItem;
      Procedure TREEPaint; Virtual;
      Procedure TREEMove(X,Y: Integer);
      Procedure TREELoad;
      Procedure TREESave;
      Procedure TREEVisible(AVisible: Boolean);

      // Visual methods
       Function EdgeZoneX(X: Integer): Boolean; Virtual;
       Function EdgeZoneY(Y: Integer): Boolean; Virtual;

      Procedure GetGraphicTools; Virtual;
      Procedure SetFadeIn(AFadeIn: Boolean);
      Procedure SetAlign(Align: TAlignment); Virtual;
      Procedure SetTextAlign(ATextAlign: TTextAlignment);
   Public
      Property FocusedItem: TGUIItem Read fFocusedItem;
      Property HoveredItem: TGUIItem Read fHoveredItem;
      Property GraphTool: TGraphManager Read fGraphTool;
      Property Theme: TTheme Read fTheme;
      Property Visible: Boolean Read fVisible Write TREEVisible;
      Property Align: TAlignment Read fAlign Write SetAlign;
      Property Caption: String Read fCaption Write fCaption;
      Property TextAlign: TTextAlignment Read fTextAlign Write SetTextAlign;

      // Constructor/Destructor/Registration
      Constructor Create(AName: String; AParent: TGUIItem;
         ALeft,ATop,AWidth,AHeight: Integer); Reintroduce;

      Destructor Destroy; Override;
      Procedure Register(AGUIItem: TGUIItem);
      Procedure SetParent(AParent: TGUIItem);
      Procedure UnRegister(AGUIItem: TGUIItem);

      // Public virtual methods concerning events
      Procedure Idle; Virtual;
      Procedure KeyDown(AKey,PreviousKey: LongWord); Virtual;
      Procedure KeyUp(AKey,PreviousKey: LongWord); Virtual;
      Procedure MouseBtnLDown(X,Y,AddKeys: Integer); Virtual;
      Procedure MouseBtnLUp(X,Y,AddKeys: Integer); Virtual;
      Procedure MouseBtnRDown(X,Y,AddKeys: Integer); Virtual;
      Procedure MouseBtnMDown(X,Y,AddKeys: Integer); Virtual;
      Procedure MouseMove(X,Y: Integer); Virtual;
      Procedure MouseWheelDown(X,Y,ADelta: Integer); Virtual;
      Procedure MouseWheelUp(X,Y,ADelta: Integer); Virtual;

   Published
      // Visual properties
      Property FrameWidth: Integer Read fFrameWidth Write fFrameWidth;
      Property BGColor: TRGBAColor Read fBGColor Write fBGColor;
      Property FGColor: TRGBAColor Read fFGColor Write fFGColor;
      Property Shape: TShape Read fShape Write fShape;
      Property Font: TFreeType2Font Read fFont Write fFont;
      Property Image: TRGBAImage Read fImage Write fImage;
      Property FadeIn: Boolean Read fFadeIn Write SetFadeIn;
      // Position properties
      Property Left: Integer Read fLeft Write fLeft;
      Property Top: Integer Read fTop Write fTop;
      Property Width: Integer Read fWidth Write fWidth;
      Property Height: Integer Read fHeight Write fHeight;

      // Event properties
      Property OnKeyDown: TKeyEvent Read fOnKeyDown Write fOnKeyDown;
      Property OnKeyUp: TKeyEvent Read fOnKeyUp Write fOnKeyUp;
      Property OnMouseBtnLDown: TMouseBtnEvent Read fOnMouseBtnLDown Write fOnMouseBtnLDown;
      Property OnMouseBtnLUp: TMouseBtnEvent Read fOnMouseBtnLUp Write fOnMouseBtnLUp;
      Property OnMouseBtnRDown: TMouseBtnEvent Read fOnMouseBtnRDown Write fOnMouseBtnRDown;
      Property OnMouseBtnMDown: TMouseBtnEvent Read fOnMouseBtnMDown Write fOnMouseBtnMDown;
      Property OnMouseMove: TMouseMoveEvent Read fOnMouseMove Write fOnMouseMove;
      Property OnMouseWheelDown: TMouseWheelEvent Read fOnMouseWheelDown Write fOnMouseWheelDown;
      Property OnMouseWheelUp: TMouseWheelEvent Read fOnMouseWheelUp Write fOnMouseWheelUp;

      Property OnPaint: TNotifyEvent Read fOnPaint Write fOnPaint;
      Property OnIdle: TNotifyEvent Read fOnIdle Write fOnIdle;
      Property OnResize: TResizeEvent Read fOnResize Write fOnResize;
   End;
//********************************************************************
   TGUIWindow = Class(TGUIItem)
   Private
      fHandle: LongWord;
      fRegion: LongWord;
      fRenderContext: LongWord;
      fCaption: String;
      fIsMainWindow: Boolean;
      fAbsolutePos: TWindowRect;
      fOldAbsPos: TWindowRect;

      Procedure GetRenderingContext(ARenderContext: Boolean);
      Procedure SetRenderingContext(ARenderContext: LongWord);

   Protected
       Function EdgeZoneX(X: Integer): Boolean; Override;
       Function EdgeZoneY(Y: Integer): Boolean; Override;

      Procedure GetGraphicTools; Override;
   Public
      Property IsMainWindow: Boolean Read fIsMainWindow;
      Property Handle: LongWord Read fHandle;
      Property RenderContext: LongWord
         Read fRenderContext Write SetRenderingContext;

      Constructor Create(AName: String; AParent: TGUIItem;
          ARenderContext: Boolean;
          AWindowClassName, ACaption: String;
          IsMainWindow: Boolean;
          ALeft,ATop,AWidth,AHeight: Integer); Reintroduce; Virtual;
      Destructor Destroy; Override;

      // Public virtual methods concerning events
      Procedure Move(X,Y: Integer); Override;
      Procedure Resize(X,Y: Integer); Override;

      Procedure KeyDown(AKey,PreviousKey: LongWord); Override;
      Procedure KeyUp(AKey,PreviousKey: LongWord); Override;
      Procedure MouseBtnLDown(X,Y,AddKeys: Integer); Override;
      Procedure MouseBtnLUp(X,Y,AddKeys: Integer); Override;
      Procedure MouseBtnRDown(X,Y,AddKeys: Integer); Override;
      Procedure MouseBtnMDown(X,Y,AddKeys: Integer); Override;
      Procedure MouseMove(X,Y: Integer); Override;
      Procedure MouseWheelDown(X,Y,ADelta: Integer); Override;
      Procedure MouseWheelUp(X,Y,ADelta: Integer); Override;

      Procedure Maximize;
      Procedure Minimize;
      Procedure Normalize;
      Procedure TREEPaint; Override;
      Procedure Show;
   End;
//********************************************************************
Function GetKeyEvent(APointer: Pointer): TKeyEvent;
Function GetMouseBtnEvent(APointer: Pointer): TMouseBtnEvent;
Function GetMouseMoveEvent(APointer: Pointer): TMouseMoveEvent;
Function GetMouseWheelEvent(APointer: Pointer): TMouseWheelEvent;
Function GetNotifyEvent(APointer: Pointer): TNotifyEvent;
Function GetResizeEvent(APointer: Pointer): TResizeEvent;
//********************************************************************
IMPLEMENTATION
//********************************************************************
//********************************************************************
//   T T H E M E
//********************************************************************
Constructor TTheme.Create;
Begin
   Inherited Create;
   fThemeDef:= TList.Create;
End;
//********************************************************************
Destructor TTheme.Destroy;
Var LocalThemeDef: PThemeDef; I: Integer;
Begin
   With fThemeDef Do If Count > 0
      Then For I:= 0 To Count-1 Do Begin
         LocalThemeDef:= Items[I];
         Dispose(LocalThemeDef);
      End;
   fThemeDef.Free;

   Inherited Destroy;
End;
//********************************************************************
Procedure TTheme.Load;
Var ThemeList: TStringList; LFileName: String; I: Integer;
Begin
   LFileName:= GetCurrentDir + '\Themes\Standard.CSV';
   If FileExists(LFileName) Then Begin
      ThemeList:= TStringList.Create;
      ThemeList.LoadFromFile(LFileName);

      With ThemeList Do If Count > 0
         Then For I:= 0 To Count-1 Do
            AddThemeClass(ThemeList[I]);

      ThemeList.Free;
   End;
End;
//********************************************************************
Procedure TTheme.Save;
Var ThemeList: TStringList; I: Integer;
Begin
   With fThemeDef Do If Count > 0 Then Begin

      ThemeList:= TStringList.Create;

      For I:= 0 To Count-1 Do
         ThemeList.Add(GetThemeClass(I));

      ThemeList.SaveToFile('/Themes/Standard.CSV');
      ThemeList.Free;
   End;
End;
//********************************************************************
Function TTheme.GetThemeCount: Integer;
Begin
   Result:= fThemeDef.Count;
End;
//********************************************************************
// Save
Function TTheme.GetThemeClass(AIndex: Integer): String;
Var LocalThemeDef: PThemeDef;
Begin
   With fThemeDef Do
      If (AIndex >= 0) And (AIndex < Count) Then Begin
         LocalThemeDef:= fThemeDef[AIndex];
         With LocalThemeDef^ Do Begin
            Result:= thClass    + ';' +
                     thColor    + ';' +
                     thShape + ';' +
                     thFont     + ';' +
                     thImage    + ';' ;
         End;
      End;
End;
//********************************************************************
// Load
Procedure TTheme.AddThemeClass(AThemeClass: String);
Var CrackList: TStringList; LocalThemeDef: PThemeDef;
Begin
   CrackList:= TStringList.Create;
   CrackUpString(AThemeClass,CrackList,';',TRUE);
   New(LocalThemeDef);

   With LocalThemeDef^ Do Begin
      thClass    := Uppercase(CrackList[0]);
      thColor    := Uppercase(CrackList[1]);
      thShape := Uppercase(CrackList[2]);
      thFont     := Uppercase(CrackList[3]);
      thImage    := Uppercase(CrackList[4]);
   End;

   fThemeDef.Add(LocalThemeDef);
   CrackList.Free;
End;
//********************************************************************
Function TTheme.GetThemeColorByIndex(AIndex: Integer): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByIndex(AIndex);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thColor
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeFontByIndex(AIndex: Integer): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByIndex(AIndex);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thFont
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeImageByIndex(AIndex: Integer): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByIndex(AIndex);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thImage
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeShapeByIndex(AIndex: Integer): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByIndex(AIndex);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thShape
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeColor(AThemeClass: String): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByName(AThemeClass);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thColor
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeShape(AThemeClass: String): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByName(AThemeClass);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thShape
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeFont(AThemeClass: String): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByName(AThemeClass);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thFont
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.GetThemeImage(AThemeClass: String): String;
Var LocalThemeDef: PThemeDef;
Begin
   LocalThemeDef:= ThemeDefByName(AThemeClass);
   If LocalThemeDef <> Nil
      Then Result:= LocalThemeDef^.thImage
      Else Result:= 'ERROR';
End;
//********************************************************************
Function TTheme.ThemeDefByIndex(AIndex: Integer): PThemeDef;
Begin
   Result:= Nil;

   If (AIndex >= 0) And (AIndex < fThemeDef.Count)
      Then Result:= fThemeDef[AIndex];
End;
//********************************************************************
Function TTheme.ThemeDefByName(AThemeClass: String): PThemeDef;
Var LocalThemeDef: PThemeDef; I: Integer; Found: Boolean;
Begin
   Result:= Nil;

   AThemeClass:= UpperCase(AThemeClass);
   I:=0;
   Found:= False;
   With fThemeDef Do While (I < Count) And (Not Found) Do Begin
      LocalThemeDef:= List[I];
      If LocalThemeDef^.thClass <> AThemeClass
         Then Inc(I) //================>
         Else Begin
            Result:= LocalThemeDef;
            Found:= TRUE;
         End;
   End;
End;
//********************************************************************
//********************************************************************
//   T G U I I T E M   -   Constructor/Destructor/Registration
//********************************************************************
Constructor TGUIItem.Create(AName: String; AParent: TGUIItem;
                  ALeft,ATop,AWidth,AHeight: Integer);
Begin
   Inherited Create(AName);

   fRoot:= Nil;
   fParent:= Nil;
   SetParent(AParent);
   fChildren:= TObjectList.Create;

   If (AParent <> Nil) And (AParent.GraphTool <> Nil)
      Then Begin
         fRoot:= AParent.fRoot;
         fGraphTool:= AParent.GraphTool;
         fTheme:= AParent.Theme;
         GetGraphicTools;
      End;

   fVisible:= True;

   fLeft   := ALeft;
   fTop    := ATop;
   fWidth  := AWidth;
   fHeight := AHeight;

   fDrawMode:= dmAll;
   fFrameState:= fsFlat;
   fFrameWidth:= 5;
   fFadeIn:= False;
   fRepeat:= False;
   fAlign:= alNone;
   SetTextAlign(taTopLeft);  
End;
//********************************************************************
Destructor TGUIItem.Destroy;
Begin
   fChildren.Free;
   Inherited Destroy;
End;
//********************************************************************
Procedure TGUIItem.Register(AGUIItem: TGUIItem);
Begin
   fChildren.Add(AGUIItem);
End;
//********************************************************************
// Set a new parent
Procedure TGUIItem.SetParent(AParent: TGUIItem);
Begin
   If Assigned(fParent)
      Then fParent.Unregister(Self);
   fParent:= AParent;
   If Assigned(fParent)
      Then fParent.Register(Self);
End;
//********************************************************************
// Deletes a GUIItem in the list
Procedure TGUIItem.Unregister(AGUIItem: TGUIItem);
Var Index: Integer;
Begin
   With fChildren Do Begin
      Index:= IndexOf(AGUIItem);
      If Index > -1
         Then Extract(AGUIItem);
   End;
End;
//********************************************************************
Procedure TGUIItem.GetGraphicTools;
Var
   ThemeClassName,
   ThemeColorName,
   ThemeFontName,
   ThemeImageName,
   ThemeShapeName: String;
Begin
   ThemeClassName:= ClassName;

   With fTheme Do Begin
      ThemeColorName:= fTheme.GetThemeColor(ThemeClassName);
      ThemeShapeName:= fTheme.GetThemeShape(ThemeClassName);
      ThemeFontName:= fTheme.GetThemeFont(ThemeClassName);
      ThemeImageName:= fTheme.GetThemeImage(ThemeClassName);
   End;

   With GraphTool Do Begin
      BGColor:= ColorByName['SolidGrey'];

      fFGColor:= ColorByName[ThemeColorName];
      If fFGColor = Nil
         Then fFGColor:= ColorByName['SolidGrey'];

      fFont:= FontByName[ThemeFontName];
      If fFont = Nil
         Then fFont:= FontByIndex[0];

      fImage:= ImageByName[ThemeImageName];
      If fImage = Nil
         Then fImage:= ImageByIndex[0];

      fShape:= ShapeByName[ThemeShapeName];
      If fShape = Nil
         Then fShape:= ShapeByName['Rectangle'];
   End;
End;
//********************************************************************
//   T G U I I T E M   -   Protected methods
//********************************************************************
Function TGUIItem.Hit(X,Y: Integer): TGUIItem;
Begin
   Result:= Nil;
   If (fVisible) And
      (X >= fLeft) And (X <= fLeft+fWidth) And
      (Y >= fTop) And (Y <= fTop+fHeight)
      Then Result:= Self;
End;
//********************************************************************
Procedure TGUIItem.Idle;
Begin
   If Assigned(fOnIdle)
      Then fOnIdle;
End;
//********************************************************************
Procedure TGUIItem.Move(X,Y: Integer);
Begin
   Inc(fLeft, X);
   Inc(fTop,  Y);
End;
//********************************************************************
Procedure TGUIItem.Paint;
Begin
   With FGColor Do
      If (fFadeIn) And (Alpha < 255)
         Then Begin
            Alpha:= Alpha+1;
            If Alpha < 255
               Then Begin
                  fRoot.fRepeat:= True;
                  Sleep(20);
               End
               Else fRoot.fRepeat:= False;
         End;

   fShape.Show(fLeft,fTop,fWidth,fHeight, fDrawMode,
        fFGColor,fImage,fFrameWidth,fFrameState);

   If Assigned(fOnPaint)
      Then fOnPaint;
End;
//********************************************************************
Procedure TGUIItem.Resize(X,Y: Integer);
Const Tolerance = 10;
Begin
   If X > (fLeft + Tolerance)
      Then fWidth:= X - fLeft + Tolerance;
   If Y > (fTop + Tolerance)
      Then fHeight:= Y - fTop + Tolerance;

   If Assigned(fOnResize)
      Then fOnResize(X,Y);
End;
//********************************************************************
// TREE-method: forces all visible children to check if hit
Function TGUIItem.TREEHit(X,Y: Integer): TGUIItem;
Var LocalChild, HitItem: TGUIItem; I: Integer;
Begin
   HitItem:= Hit(X,Y);

   If HitItem = Self Then Begin

      // Search from the last to the first
      // (just the other way round they are drawn)
      With fChildren Do If Count > 0 Then Begin
         I:=Count-1;
         While (I >= 0) And (HitItem = Self) Do Begin
            LocalChild:= List[I];
            LocalChild:= LocalChild.TREEHit(X,Y);
            If LocalChild <> Nil
               Then HitItem:= LocalChild
               Else Dec(I);
         End;
      End;
   End;
   Result:= HitItem;
End;
//********************************************************************
// TREE-method: forces all children to find if the passed name
// does not fit
Function TGUIItem.TREEItemByName(AName: String): TGUIItem;
Var LocalChild, FoundItem: TGUIItem; I: Integer;
Begin
   FoundItem:= Nil;
   AName:= UpperCase(AName);

   If Name = AName
      Then FoundItem:= Self
      Else With fChildren Do If Count > 0 Then Begin
         I:= 0;
         While (I < Count) And (FoundItem = Nil) Do Begin
            LocalChild:= List[I];
            LocalChild:= LocalChild.TREEItemByName(AName);
            If LocalChild <> Nil
               Then FoundItem:= LocalChild
               Else Inc(I); //=================>
         End;
      End;
   Result:= FoundItem;
End;
//********************************************************************
// TREE-method: forces all children to Paint
Procedure TGUIItem.TREEPaint;
Var LocalChild: TGUIItem; I: Integer;
Begin
   Paint;

   With fChildren Do If Count > 0
      Then For I:=0 To Count-1 Do Begin
         LocalChild:= List[I];
         LocalChild.TREEPaint;
      End;
End;
//********************************************************************
// TREE-method: forces all children to Move
Procedure TGUIItem.TREEMove(X,Y: Integer);
Var LocalChild: TGUIItem; I: Integer;
Begin
   Move(X,Y);

   With fChildren Do If Count > 0
      Then For I:=0 To Count-1 Do Begin
         LocalChild:= List[I];
         LocalChild.TREEMove(X,Y);
      End;
End;
//********************************************************************
// TREE-method: forces all children to set their visibility
Procedure TGUIItem.TREEVisible(AVisible: Boolean);
Var LocalChild: TGUIItem; I: Integer;
Begin
   fVisible:= AVisible;

   With fChildren Do If Count > 0
      Then For I:=0 To Count-1 Do Begin
         LocalChild:= List[I];
         LocalChild.TREEVisible(AVisible);
      End;
End;
//********************************************************************
// TREE-method: forces all children to load
Procedure TGUIItem.TREELoad;
Var LocalChild: TGUIItem; I: Integer;
Begin
   //Loading

   With fChildren Do If Count > 0
      Then For I:=0 To Count-1 Do Begin
         LocalChild:= List[I];
         LocalChild.TREELoad;
      End;
End;
//********************************************************************
// TREE-method: forces all children to save
Procedure TGUIItem.TREESave;
Var LocalChild: TGUIItem; I: Integer;
Begin
   //Saving

   With fChildren Do If Count > 0
      Then For I:=0 To Count-1 Do Begin
         LocalChild:= List[I];
         LocalChild.TREESave;
      End;
End;
//********************************************************************
Procedure TGUIItem.SetAlign(Align: TAlignment);
Const Margin = 2;
Begin
   fAlign:= Align;     

   If fParent <> Nil Then Case fAlign Of
      alAll    : Begin
                    fLeft:=fParent.Left+Margin;
                    fTop:=fParent.Top+Margin;
                    fWidth:=fParent.Width-2*Margin;
                    fHeight:=fParent.Height-2*Margin;
                 End;
      alLeft   : Begin
                    fLeft:=fParent.Left+Margin;
                    fTop:=fParent.Top+Margin;
                    fHeight:=fParent.Height-2*Margin;
                 End;
      alRight  : Begin;
                    fLeft:=fParent.Left+fParent.Width-fWidth-Margin;
                    fTop:=fParent.Top+Margin;
                    fHeight:=fParent.Height-2*Margin;
                 End;
      alTop    : Begin;
                    fLeft:=fParent.Left+Margin;
                    fTop:=fParent.Top+Margin;
                    fWidth:=fParent.Width-2*Margin;
                 End;
      alBottom : Begin;
                    fLeft:=fParent.Left+Margin;
                    fTop:=fParent.Top+fParent.Height-fHeight-Margin;
                    fWidth:=fParent.Width-2*Margin;
                 End;
   End;
End;
//********************************************************************
Procedure TGUIItem.SetTextAlign(ATextAlign: TTextAlignment);
Const Margin = 3;
Var TextW: Integer;
Procedure XLeft;
Begin
   fTextPos.X:= Left+Margin;
End;
Procedure XRight;
Begin
   With fTextPos Do Begin
      X:= fLeft+fWidth-Font.TextWidth(fCaption)-Margin;
      If X < 0 Then X:= 0;
   End;
End;
Procedure XCenter;
Begin
   With fTextPos Do Begin
      X:= fLeft+(fWidth Div 2)-(Font.TextWidth(fCaption) Div 2)
          - Margin Div 2;
      If X < 0 Then X:= 0;
   End;
End;
Procedure YTop;
Begin
   fTextPos.Y:= Top+Margin;
End;
Procedure YBottom;
Begin
   With fTextPos Do Begin
      Y:= fTop+fHeight-Font.TextHeight(fCaption)-Margin;
      If Y < 0 Then Y:= 0;
   End;
End;
Procedure YCenter;
Begin
   With fTextPos Do Begin
      Y:= fTop+(fHeight Div 2)-(Font.TextHeight(fCaption) Div 2)
          - Margin Div 2;
      If Y < 0 Then Y:= 0;
   End;
End;
Begin
   fTextAlign:= ATextAlign;
   TextW:= Font.TextWidth(fCaption);

   With fTextPos Do Case fTextAlign Of
      taTopLeft   : Begin XLeft; YTop; End;
      taTopRight  : Begin XRight; YTop; End;
      taTopCenter : Begin XCenter; YTop; End;
      taBottomLeft   : Begin XLeft; YBottom; End;
      taBottomRight  : Begin XRight; YBottom; End;
      taBottomCenter : Begin XCenter; YBottom; End;
      taMidLeft   : Begin XLeft; YCenter; End;
      taMidRight  : Begin XRight; YCenter; End;
      taMidCenter : Begin XCenter; YCenter; End;
   End;
End;
//********************************************************************
Procedure TGUIItem.SetFadeIn(AFadeIn: Boolean);
Begin
   fFadeIn:= AFadeIn;
   If fFadeIn
      Then fFGColor.Alpha:=0
      Else fFGColor.Alpha:=255;
End;
//********************************************************************
Function TGUIItem.EdgeZoneX(X: Integer): Boolean;
Const EdgeWidth = 20;
Begin
   Result:= False;
   If (Abs(fLeft + fWidth - X)  < EdgeWidth)
      Then Result:= True;
End;
//********************************************************************
Function TGUIItem.EdgeZoneY(Y: Integer): Boolean;
Const EdgeWidth = 20;
Begin
   Result:= False;
   If (Abs(fTop + fHeight - Y) < EdgeWidth)
      Then Result:= True;
End;
//********************************************************************
//   T G U I I T E M  -  Public virtual methods concerning events
//********************************************************************
Procedure TGUIItem.KeyDown(AKey,PreviousKey: LongWord);
Begin
   If Assigned(fOnKeyDown)
      Then fOnKeyDown(AKey,PreviousKey);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.KeyUp(AKey,PreviousKey: LongWord);
Begin
   If Assigned(fOnKeyUp)
      Then fOnKeyUp(AKey,PreviousKey);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseBtnLDown(X,Y,AddKeys: Integer);
Begin
   If Assigned(fOnMouseBtnLDown)
      Then fOnMouseBtnLDown(X,Y,AddKeys);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseBtnLUp(X,Y,AddKeys: Integer);
Begin
   If Assigned(fOnMouseBtnLUp)
      Then fOnMouseBtnLUp(X,Y,AddKeys);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseBtnRDown(X,Y,AddKeys: Integer);
Begin
   If Assigned(fOnMouseBtnMDown)
      Then fOnMouseBtnMDown(X,Y,AddKeys);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseBtnMDown(X,Y,AddKeys: Integer);
Begin
   If Assigned(fOnMouseBtnMDown)
      Then fOnMouseBtnMDown(X,Y,AddKeys);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseMove(X,Y: Integer);
Begin
   If Assigned(fOnMouseMove)
      Then fOnMouseMove(X,Y);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseWheelDown(X,Y,ADelta: Integer);
Begin
   If Assigned(fOnMouseWheelDown)
      Then fOnMouseWheelDown(X,Y,ADelta);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
Procedure TGUIItem.MouseWheelUp(X,Y,ADelta: Integer);
Begin
   If Assigned(fOnMouseWheelUp)
      Then fOnMouseWheelUp(X,Y,ADelta);
   If Assigned(fRoot)
      Then fRoot.TREEPaint;
End;
//********************************************************************
//********************************************************************
//   T G U I W I N D O W
//********************************************************************
Constructor TGUIWindow.Create(AName: String; AParent: TGUIItem;
          ARenderContext: Boolean;
          AWindowClassName, ACaption: String;
          IsMainWindow: Boolean;
          ALeft,ATop,AWidth,AHeight: Integer);
Begin
   Inherited Create(AName,AParent,0,0,AWidth,AHeight);

   With fAbsolutePos Do Begin
      Left  := ALeft;
      Top   := ATop;
      Right := ALeft+AWidth;
      Bottom:= ATop+AHeight;
   End;
   fOldAbsPos:= fAbsolutePos;

   // Attention - these variables belong to
   // TGUIItem/TGUIItem!
   fRoot:= Self;        // TGUIItem, protected
   fFocusedItem:= Nil;  // TGUIItem, protected
   fHoveredItem:= Nil;  // TGUIItem, protected
   fDragMode:= 0;       // TGUIItem, protected
   //--------------------------------------------

   fCaption:= ACaption;
   fIsMainWindow:= IsMainWindow;
   fDrawMode:= dmFrameOnly;
   fFrameState:= fsRaised;

   With fAbsolutePos Do
      fHandle:= WindowCreate(AWindowClassName,fCaption,
                Left,Top,Right-Left,Bottom-Top);
   //fRegion:= WindowRegion(fHandle,Width,Height);

   GetRenderingContext(ARenderContext);
   MouseImageSet(fHandle,miArrow,0,0);
End;
//********************************************************************
Destructor TGUIWindow.Destroy;
Begin
   // Inform Graphic Library
   fGraphTool.Free;
   RenderContextFree(fHandle,fRenderContext);

   // Inform Operating System
   WindowDestroy(fHandle,fRegion);

   fTheme.Free;

   Inherited Destroy;
End;
//********************************************************************
Procedure TGUIWindow.GetGraphicTools;
Var
   LName: String; I: Integer;
Begin
   fTheme:= Nil;
   fTheme:= TTheme.Create;
   fTheme.Load;

   fGraphTool:= TGraphManager.Create(fHandle);
   OGLBGColor:= SOLIDLGREY;  // Color: GL-BackGround
   GraphTool.Initialize;

   // Load Theme: charge GraphTool to load Fonts and Textures
   // (GraphTool.Register checks for duplicates)
   With fTheme Do If Count > 0 Then For I:= 0 To Count-1 Do Begin

      // Colors, Symbols and Shapes are set up by GraphTool itself

      // Load font
      LName:= fTheme.FontByIndex[I];
      If GraphTool.FontByName[LName] = Nil
         Then TFreeType2Font.Create(LName,GraphTool);

      // Load image (texture)
      LName:= fTheme.ImageByIndex[I];
      If GraphTool.ImageByName[LName] = Nil
         Then TRGBAImage.Create(LName,GraphTool);
   End;

   Inherited GetGraphicTools;
End;
//********************************************************************
Procedure TGUIWindow.GetRenderingContext(ARenderContext: Boolean);
Begin
   If ARenderContext
      Then Begin
         fRenderContext:= RenderContextCreate(fHandle);
         SetRenderingContext(fRenderContext);
      End
      Else fRenderContext:= 0;

   WindowShow(fHandle, TRUE);
End;
//********************************************************************
Procedure TGUIWindow.SetRenderingContext(ARenderContext: LongWord);
Begin
   fRenderContext:= ARenderContext;
   Resize(Width,Height);
   WindowResize(fHandle,Width,Height);
   GetGraphicTools;
End;
//********************************************************************
Procedure TGUIWindow.Maximize;
Begin

   With fAbsolutePos Do Begin
      Move(-Left,-Top);
      WindowMove(fHandle,Left,Top);
   End;
   Resize(1024,768);
   WindowResize(fHandle,Width,Height);
End;
//********************************************************************
Procedure TGUIWindow.Minimize;
Begin
   WindowMinimize(fHandle);
End;
//********************************************************************
Procedure TGUIWindow.Normalize;
Begin
   Move(fOldAbsPos.Left-fAbsolutePos.Left,
         fOldAbsPos.Top-fAbsolutePos.Top);
   WindowMove(fHandle,fAbsolutePos.Left,fAbsolutePos.Top);

   With fOldAbsPos Do
      Resize(Right-Left,Bottom-Top);
   WindowResize(fHandle,Width,Height);
End;
//********************************************************************
Procedure TGUIWindow.Show;
Begin
   If Visible Then Begin
      //GraphTool.StartOrtho;
      Inherited TREEPAINT;
      //GraphTool.StopOrtho;
   End;
End;
//********************************************************************
Procedure TGUIWindow.TREEPaint;
Begin
   If Visible Then Begin
      If RenderContextActivate(fHandle,fRenderContext)
         Then Repeat
            GraphTool.PaintBegin;
            Show;
            GraphTool.PaintEnd;
         Until fRepeat = False;
         RenderContextDeactivate;
   End;
End;
//********************************************************************
//   T G U I W I N D O W   -   Protected virtual methods
//********************************************************************
Function TGUIWindow.EdgeZoneX(X: Integer): Boolean;
Const EdgeWidth = 20;
Begin
   Result:= False;
   If Abs(Width - X)  < EdgeWidth
      Then Result:= True;
End;
//********************************************************************
Function TGUIWindow.EdgeZoneY(Y: Integer): Boolean;
Const EdgeWidth = 20;
Begin
   Result:= False;
   If Abs(Height - Y) < EdgeWidth
      Then Result:= True;
End;
//********************************************************************
Procedure TGUIWindow.Resize(X,Y: Integer);
Const Tolerance = 10;
Begin
   If X > (fAbsolutePos.Left + Tolerance)
      Then Width:= X;
   If Y > (fAbsolutePos.Top + Tolerance)
      Then Height:= Y;

   // Inform Graphic Library
   If RenderContextActivate(fHandle,fRenderContext)
      Then fGraphTool.Resize(Width,Height);

   If Assigned(OnResize)
      Then OnResize(X,Y);
End;
//********************************************************************
Procedure TGUIWindow.Move(X,Y: Integer);
Begin
   With fAbsolutePos Do Begin
      Inc(Left, X);
      Inc(Top,  Y);

      WindowMove(fHandle,Left,Top);
   End;
End;
//********************************************************************
//   T G U I W I N D O W   -  Public virtual methods concerning events
//********************************************************************
Procedure TGUIWindow.KeyDown(AKey,PreviousKey: LongWord);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.KeyDown(AKey,PreviousKey);
   Inherited KeyDown(AKey,PreviousKey);
End;
//********************************************************************
Procedure TGUIWindow.KeyUp(AKey,PreviousKey: LongWord);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.KeyUp(AKey,PreviousKey);
   Inherited KeyUp(AKey,PreviousKey);
End;
//********************************************************************
Procedure TGUIWindow.MouseBtnLDown(X,Y,AddKeys: Integer);
Begin
   fFocusedItem:= TREEHit(X,Y); // Get focus on Mouseklick

   If (fFocusedItem <> Nil) Then Begin
      If(fFocusedItem <> Self)
         Then fFocusedItem.MouseBtnLDown(X,Y,AddKeys)
         Else If AddKeys = KEY_CONTROL Then Begin
            MouseCaptureRelease;
            If (EdgeZoneX(X) Or EdgeZoneY(Y))
               Then Begin
                  fDragMode:= 1;  // Sizing
                  MouseImageSet(fHandle,miSizing,X,Y);
                  fFrameWidth:= 0;
                  WindowResizeByMouse(fHandle);
               End
               Else Begin
                  fDragMode:= 2;  // Moving
                  MouseImageSet(fHandle,miMoving,X,Y);
                  WindowMoveByMouse(fHandle);
               End;
         End;
   End;

   Inherited MouseBtnLDown(X,Y,AddKeys);
End;
//********************************************************************
Procedure TGUIWindow.MouseBtnLUp(X,Y,AddKeys: Integer);
Begin
   If (fFocusedItem <> Nil) Then Begin
      If (fFocusedItem <> Self)
         Then fFocusedItem.MouseBtnLUp(X,Y,AddKeys);
   End;

   Inherited MouseBtnLUp(X,Y,AddKeys);
End;
//********************************************************************
Procedure TGUIWindow.MouseBtnRDown(X,Y,AddKeys: Integer);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.MouseBtnRDown(X,Y,AddKeys);
   Inherited MouseBtnRDown(X,Y,AddKeys);
End;
//********************************************************************
Procedure TGUIWindow.MouseBtnMDown(X,Y,AddKeys: Integer);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.MouseBtnMDown(X,Y,AddKeys);
   Inherited MouseBtnMDown(X,Y,AddKeys);
End;
//********************************************************************
Procedure TGUIWindow.MouseMove(X,Y: Integer);
Begin
   // Calls user proc + TREEPaint for HoveredItem
   If (fHoveredItem <> Nil) And (HoveredItem <> Self)
      Then fHoveredItem.MouseMove(X,Y);

   // Set HoveredItem
   // IMPORTANT: HoveredItem should be set AFTER calling
   // fHoveredItem.MouseMove, otherwise bad hittest
   fHoveredItem:= TREEHit(X,Y);

   // Move/Resize GUIWindow
   If (fFocusedItem <> Nil) Then Begin
      If (fFocusedItem <> Self)
         Then fFocusedItem.MouseMove(X,Y)
         // This is called after dragging the window by WM_SYSCOMMAND
         // so HERE is the place for turning the DragMode OFF
         Else If fDragMode <> 0
            Then Begin  // Sizing
               fAbsolutePos:= WindowRect(fHandle);
               If fDragMode = 1
                  Then With fAbsolutePos Do Begin
                     Resize(Right-Left,Bottom-Top);
                     fFrameWidth:= 5;
                  End;
               fDragMode:= 0;
               MouseImageSet(fHandle,miArrow,X,Y);
            End;
   End;
   Inherited MouseMove(X,Y);
End;
//********************************************************************
Procedure TGUIWindow.MouseWheelDown(X,Y,ADelta: Integer);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.MouseWheelDown(X,Y,ADelta);
   Inherited MouseWheelDown(X,Y,ADelta);
End;
//********************************************************************
Procedure TGUIWindow.MouseWheelUp(X,Y,ADelta: Integer);
Begin
   If (fFocusedItem <> Nil) And (fFocusedItem <> Self)
      Then fFocusedItem.MouseWheelUp(X,Y,ADelta);
   Inherited MouseWheelUp(X,Y,ADelta);
End;
//********************************************************************
//********************************************************************
//   P R O C E D U R E S
//********************************************************************
Function GetNotifyEvent(APointer: Pointer): TNotifyEvent;
Var
   Event: TNotifyEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
Function GetKeyEvent(APointer: Pointer): TKeyEvent;
Var
   Event: TKeyEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
Function GetMouseBtnEvent(APointer: Pointer): TMouseBtnEvent;
Var
   Event: TMouseBtnEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
Function GetMouseMoveEvent(APointer: Pointer): TMouseMoveEvent;
Var
   Event: TMouseMoveEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
Function GetMouseWheelEvent(APointer: Pointer): TMouseWheelEvent;
Var
   Event: TMouseWheelEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
Function GetResizeEvent(APointer: Pointer): TResizeEvent;
Var
   Event: TResizeEvent;
Begin
   TMethod(Event).Code:= APointer;
   TMethod(Event).Data:= Nil;
   Result:= Event;
End;
//********************************************************************
//********************************************************************
End.

