//********************************************************************
UNIT GraphicSupport;
//********************************************************************
INTERFACE

USES
   SysUtils,Classes,Contnrs,Math,dglOpenGl,FreeTypeH,
   ImageLoader,OSSupport;
//********************************************************************
TYPE
   TLumenPixel = Record Lumen,Alpha: Byte; End;
   PLumenPixel = ^TLumenPixel;
   TMonoPixel = Byte;
   PMonoPixel = PByte;
   TPoint2D = Record X,Y: Integer; End;
   RGBAColor = Record Red,Green,Blue,Alpha: Byte; End;
   TBGRPixel = Packed Record Blue,Green,Red: Byte; End;
   TRGBAPixel = Packed Record Red,Green,Blue,Alpha: Byte; End;
   PRGBAPixel = ^TRGBAPixel;
//********************************************************************
CONST
   SOLIDWHITE:  RGBAColor = (Red:255; Green:255; Blue:255; Alpha:255);
   TRANSWHITE:  RGBAColor = (Red:255; Green:255; Blue:255; Alpha:128);
   SOLIDBLACK:  RGBAColor = (Red:  0; Green:  0; Blue:  0; Alpha:255);
   SOLIDGREY :  RGBAColor = (Red:128; Green:128; Blue:128; Alpha:255);
   SOLIDLGREY:  RGBAColor = (Red:200; Green:200; Blue:200; Alpha:255);
   SOLIDRED  :  RGBAColor = (Red:255; Green:  0; Blue:  0; Alpha:255);
   SOLIDGREEN:  RGBAColor = (Red:  0; Green:255; Blue:  0; Alpha:255);
   SOLIDBLUE :  RGBAColor = (Red:  0; Green:  0; Blue:255; Alpha:255);

   //SOLIDOLIVE:  RGBAColor = (Red:135; Green:135; Blue:  0; Alpha:255);

   PiDiv180 = PI/180;
   PiDiv360 = PI/360;
//********************************************************************
TYPE
   TDrawMode = (dmAll, dmShapeOnly, dmFrameOnly);
   TFrameState = (fsFlat, fsRaised, fsLowered);
   TVertex = Record PtX,PtY: Single; End;
   TTexVertex = Record TexX,TexY,PtX,PtY: Single; End;
   PTexVertex = ^TTexVertex;
   TRGBAColor = Class;
   TRGBAImage = Class;
//********************************************************************
   TIDObject = Class(TPersistent)
   Private
      fName: String;
      Procedure SetName(AName: String);
   Public
      Constructor Create(AName: String); Reintroduce;
   Published
      // Properties for loading/saving
      Property Name: String Read fName Write SetName;
   End;
//********************************************************************
   TCustomOwner = Class(TObject)
   Public
      Function Register(AIDObject: TIDObject): Boolean; Virtual; Abstract;
   End;
//********************************************************************
   TIDOwnedObject = Class(TIDObject)
   Private
      fRegistered: Boolean;
   Public
      Property Registered: Boolean Read fRegistered;
      Constructor Create(AName: String;
                  AOwner: TCustomOwner); Reintroduce; Virtual;
   End;
//********************************************************************
   TRenderer = Class(TIDOwnedObject)
   Private
      Function EllipsePoint(Angle: Integer): TVertex;
      Procedure OrthoSandBoxON;
      Procedure OrthoSandBoxOFF;
      Procedure Transform(ALeft,ATop,AWidth,AHeight: Integer);
   End;
//********************************************************************
   TSymbol = Class(TRenderer)
   Private
      fSymbol: LongWord;
      Procedure Build; Virtual; Abstract;
   Public
      Constructor Create(AName: String;
                        AOwner: TCustomOwner); Override;
      Destructor Destroy; Override;

      Procedure Show(ALeft,ATop,AWidth,AHeight: Integer;
         AColor: RGBAColor; ASpecification: Integer); Virtual;
   End;
//********************************************************************
   TArrow = Class(TSymbol)
   Private
      Procedure Build; Override;
   Public
      Procedure Show(ALeft,ATop,AWidth,AHeight: Integer;
         AColor: RGBAColor; ASpecification: Integer); Override;
   End;
//********************************************************************
   TCross = Class(TSymbol)
   Private
      Procedure Build; Override;
   End;
//********************************************************************
   THook = Class(TSymbol)
   Private
      Procedure Build; Override;
   End;
//********************************************************************
   TSpot = Class(TSymbol)
   Private
      Procedure Build; Override;
   End;
//********************************************************************
// Attention! Shapes with wholes in it can't be handled at the moment
   TShape = Class(TRenderer)
   Private
      fSilhouette: TList;
      fTriangles: TList;
      fShape: LongWord;
      fFrame: TList;

      fFileName: String;

      // General static methods
      Procedure AddVertex(ATexX,ATexY,AX,AY: Single);
      Procedure BuildShape;
      Procedure BuildFrame;
      Procedure ClearLists;
      Procedure ClearFrame;
       Function Load: Boolean;
      Procedure Triangulate;
   Protected
      // Specific methods
      Procedure GenSilhouette; Virtual; Abstract;
   Public
      Constructor Create(AName: String;
                        AOwner: TCustomOwner); Override;
      Destructor Destroy; Override;

      Procedure Show(ALeft,ATop,AWidth,AHeight: Integer;
         ADrawMode: TDrawMode;
         AColor: TRGBAColor; AImage: TRGBAImage;
         AFrameWidth: Integer; AFrameState: TFrameState);
   End;
//********************************************************************
   TEllipse = Class(TShape)
   Protected
      Procedure GenSilhouette; Override;
   End;
//********************************************************************
   THexagon = Class(TShape)
   Protected
      Procedure GenSilhouette; Override;
   End;
//********************************************************************
   TRectangle = Class(TShape)
   Protected
      Procedure GenSilhouette; Override;
   End;
//********************************************************************
   TRoundRect = Class(TShape)
   Private
      fEllipseX,fEllipseY: Single;
   Protected
      Procedure GenSilhouette; Override;
      Procedure SetEllipseParams; Virtual;
   Public
      Constructor Create(AName: String;
                        AOwner: TCustomOwner); Override;
   End;
//********************************************************************
   TRoundRectHoriz = Class(TRoundRect)
   Protected
      Procedure SetEllipseParams; Override;
   End;
//********************************************************************
   TRoundRectVert = Class(TRoundRect)
   Protected
      Procedure SetEllipseParams; Override;
   End;
//********************************************************************
   TRGBAColor = Class(TIDOwnedObject)
   Private
      fBaseColor: RGBAColor;
      fShownColor: RGBAColor;

       Function GetColor: LongWord;
      Procedure SetColor(AColor: LongWord);
      Procedure SetBaseColor(ARGBAColor: RGBAColor);
      Procedure SetGradient(AGradient: Single);
   Public
      Property RGBAColor: RGBAColor Read fShownColor Write SetBaseColor;
      Property Red: Byte Read fBaseColor.Red Write fBaseColor.Red;
      Property Green: Byte Read fBaseColor.Green Write fBaseColor.Green;
      Property Blue: Byte Read fBaseColor.Blue Write fBaseColor.Blue;
      Property Alpha: Byte Read fBaseColor.Alpha Write fBaseColor.Alpha;
      Property Gradient: Single Write SetGradient;

      Constructor Create(AName: String; AOwner: TCustomOwner;
                    ARGBAColor: RGBAColor); Reintroduce;
      Procedure Reset;
      Procedure Show;
   Published
      // Properties for loading/saving
      Property BaseColor: LongWord
         Read GetColor Write SetColor;
   End;
//********************************************************************
   TRGBAImage = Class(TIDOwnedObject)
   Private
      fWidth,fHeight: Integer;
      fRGBAPixelMap: Pointer;  // = Map of TRGBAPixel
      fTexture: LongWord;
      fAlphaColor: RGBAColor;
      fFlip: Boolean;
      fFileFormat: TImageFileFormat;  // see Unit GraphicSupport
      fFileName: String;

      Procedure Expand24To32Bit(AlphaColor: RGBAColor);
      Procedure FlipHoriz;
      Procedure GenerateTexture;

      Function GetColor: LongWord;
      Procedure SetColor(AColor: LongWord);

      Function ImportBMP_BGR: Boolean;
      Function ImportTGA_BGRA: Boolean;
   Public
      Property AlphaColor: RGBAColor Read fAlphaColor Write fAlphaColor;
      Property Texture: LongWord Read fTexture Write fTexture;

      Constructor Create(AName: String;
                        AOwner: TCustomOwner); Override;
      Destructor Destroy; Override;
      Function BuildImage: Boolean;
      Procedure RGBToGrayScale;
      Procedure Show(Visible: Boolean);
   Published
      // Properties for loading/saving
      Property Color: LongWord Read GetColor Write SetColor;
      Property FileName: String Read fFileName Write fFileName;
      Property ImageFileFormat: TImageFileFormat
         Read fFileFormat Write fFileFormat;
      Property Flip: Boolean Read fFlip Write fFlip;
   End;
//********************************************************************
   TFreeType2Font = Class(TIDOwnedObject)
   Private
      fBitmap16bpp: PLumenPixel;
      fFreeTypeLib: PFT_Library;
      fGlyphFace: PFT_Face;

      fColor: RGBAColor;
      fLetterBase: LongWord;
      fTexture: LongWord;
      fTextureWidth,fTextureHeight: Integer;
      fLetterWidth: Array[0..255] Of Single;
      fLetterHeightMax,fLetterWidthMax: Word;
      fMargin: Integer;
      fFileName: String;
      fSize: Integer;
      fStyle: Integer;

      Procedure BuildDisplayLists;
      Procedure Copy8bppTo16bpp(Source: PMonoPixel;
         Dest: PLumenPixel;SourceX,SourceY,SourceWidth,
         SourceHeight,DestX,DestY,DestWidth,DestHeight: Integer);
       Function GenerateBitmap: Boolean;
      Procedure DrawLettersOnBitmap;
      Procedure GenerateTexture;
      Procedure GetFontMetrics;

      Function GetColor: LongWord;
      Procedure SetColor(AColor: LongWord);
      Procedure SetSize(ASize: Integer);
   Public
      Property RGBAColor: RGBAColor Read fColor Write fColor;

      Constructor Create(AName: String;
                        AOwner: TCustomOwner); Override;
      Destructor Destroy; Override;
      Function BuildFont: Boolean;
      Procedure Show(XPos,YPos: Integer; AString: String);
      //Procedure ShowSelect(XPos,YPos,SelStart,SelEnd: Integer;
      //                AString: String);
      Function TextHeight(AString: String): Word;
      Function TextWidth(AString: String): LongWord;
   Published
      Property FileName: String Read fFileName Write fFileName;
      Property Color: LongWord Read GetColor Write SetColor;
      Property Size: Integer Read fSize Write SetSize;
      Property Style: Integer Read fStyle Write fStyle;
   End;
//********************************************************************
   TGraphManager = Class(TCustomOwner)
   Private
      fWindow: Integer;

      fColors: TObjectList;
      fShapes: TObjectList;
      fFonts: TObjectList;
      fImages: TObjectList;
      fSymbols: TObjectList;

      Function GetColorByIndex(AIndex: Integer): TRGBAColor;
      Function GetFontByIndex(AIndex: Integer): TFreeType2Font;
      Function GetImageByIndex(AIndex: Integer): TRGBAImage;
      Function GetShapeByIndex(AIndex: Integer): TShape;
      Function GetSymbolByIndex(AIndex: Integer): TSymbol;

      Function GetColorByName(AName: String): TRGBAColor;
      Function GetFontByName(AName: String): TFreeType2Font;
      Function GetImageByName(AName: String): TRGBAImage;
      Function GetShapeByName(AName: String): TShape;
      Function GetSymbolByName(AName: String): TSymbol;

      Function GetItemByName(AName: String; AList: TObjectList): TIDObject;
      Function ValidIndex(AIndex: Integer; AList: TObjectList): Boolean;

   Public
      Property ColorByIndex[Index: Integer]: TRGBAColor
         Read GetColorByIndex;
      Property FontByIndex[Index: Integer]: TFreeType2Font
         Read GetFontByIndex;
      Property ImageByIndex[Index: Integer]: TRGBAImage
         Read GetImageByIndex;
      Property ShapeByIndex[Index: Integer]: TShape
         Read GetShapeByIndex;
      Property SymbolByIndex[Index: Integer]: TSymbol
         Read GetSymbolByIndex;

      Property ColorByName[Index: String]: TRGBAColor
         Read GetColorByName;
      Property FontByName[Index: String]: TFreeType2Font
         Read GetFontByName;
      Property ImageByName[Index: String]: TRGBAImage
         Read GetImageByName;
      Property ShapeByName[Index: String]: TShape
         Read GetShapeByName;
      Property SymbolByName[Index: String]: TSymbol
         Read GetSymbolByName;

      Constructor Create(AWindow: LongWord); Reintroduce;
      Destructor Destroy; Override;
      Procedure Initialize;
       Function Register(AIDObject: TIDObject): Boolean; Override;
      Procedure Resize(AWidth,AHeight: Integer);
      Procedure PaintBegin;
      Procedure PaintEnd;
      Procedure SetLineWidth(ALineWidth: Integer);

   End;
//********************************************************************
Function InitializeGraphLib: Boolean;
//********************************************************************
Var OGLBGColor: RGBAColor;
   ErrorString: String;
//********************************************************************
IMPLEMENTATION
//********************************************************************
Function InitializeGraphLib: Boolean;
Begin
   Result:= InitOpenGL;
End;
//********************************************************************
Function RGBAToCardinal(ARGBAColor: RGBAColor): LongWord;
Begin
   With ARGBAColor Do
      Result:= LongWord
         ((Alpha Shl 24) + (Blue Shl 16) + (Green Shl 8) + Red);
End;
//********************************************************************
Function CardinalToRGBA(AColor: LongWord): RGBAColor;
Begin
   With Result Do Begin
      Red   :=  (AColor And $000000FF);
      Green := ((AColor And $0000FF00) Shr 8);
      Blue  := ((AColor And $00FF0000) Shr 16);
      Alpha := ((AColor And $FF000000) Shr 24);
   End;
End;
//********************************************************************
//********************************************************************
//   T G R A P H M A N A G E R
//********************************************************************
Constructor TGraphManager.Create(AWindow: LongWord);
Begin
   Inherited Create;
   fWindow:= AWindow;

   // Set up basic colors --------------------------------------------
   fColors:= TObjectList.Create;
   TRGBAColor.Create('SOLIDWHITE',  Self, SOLIDWHITE);
   TRGBAColor.Create('TRANSWHITE',  Self, TRANSWHITE);
   TRGBAColor.Create('SOLIDBLACK',  Self, SOLIDBLACK);
   TRGBAColor.Create('SOLIDGREY',   Self, SOLIDGREY);
   TRGBAColor.Create('SOLIDRED',    Self, SOLIDRED);
   TRGBAColor.Create('SOLIDGREEN',  Self, SOLIDGREEN);
   TRGBAColor.Create('SOLIDBLUE',   Self, SOLIDBLUE);
   TRGBAColor.Create('SOLIDLGREY',  Self, SOLIDLGREY);

   // Set up basic font ----------------------------------------------
   fFonts:= TObjectList.Create;
   TFreeType2Font.Create('timr46w',Self);

   // Set up basic images --------------------------------------------
   fImages:= TObjectList.Create;
   TRGBAImage.Create('DummyImage',Self);
   TRGBAImage.Create('FIGURES64X64',Self);

   // Set up ALL shapes ----------------------------------------------
   fShapes:= TObjectList.Create;
   TEllipse.Create  ('ELLIPSE',   Self);
   THexagon.Create  ('HEXAGON',   Self);
   TRectangle.Create('RECTANGLE', Self);
   TRoundRect.Create('ROUNDRECT', Self);
   TRoundRectHoriz.Create('ROUNDRECTHORIZ', Self);
   TRoundRectVert.Create('ROUNDRECTVERT', Self);

   // Set up ALL symbols ---------------------------------------------
   fSymbols:= TObjectList.Create;
   TArrow.Create('ARROW',Self);
   TCross.Create('CROSS',Self);
   THook.Create('HOOK',Self);
   TSpot.Create('SPOT',Self);
End;
//********************************************************************
Destructor TGraphManager.Destroy;
Begin
   fSymbols.Free;
   fImages.Free;
   fFonts.Free;
   fShapes.Free;
   fColors.Free;

   Inherited Destroy;
End;
//********************************************************************
Function TGraphManager.Register(AIDObject: TIDObject): Boolean;
Var
   LocalIDObject: TIDObject;
Begin
   Result:= False;

   If Assigned(AIDObject) Then Begin
      //----------------------------------------------------
      If AIDObject Is TFreeType2Font
         Then Begin
            LocalIDObject:= FontByName[AIDObject.Name];
            If LocalIDObject = Nil
               Then Begin
                  fFonts.Add(AIDObject);
                  Result:= True;
               End;
         End;
      //----------------------------------------------------
      If AIDObject Is TRGBAColor
         Then Begin
            LocalIDObject:= ColorByName[AIDObject.Name];
            If LocalIDObject = Nil
               Then Begin
                  fColors.Add(AIDObject);
                  Result:= True;
               End;
         End;
      //----------------------------------------------------
      If AIDObject Is TRGBAImage
         Then Begin
            LocalIDObject:= ImageByName[AIDObject.Name];
            If LocalIDObject = Nil
               Then Begin
                  fImages.Add(AIDObject);
                  Result:= True;
               End;
         End;
      //----------------------------------------------------
      If AIDObject.InheritsFrom(TSymbol)
         Then Begin
            LocalIDObject:= SymbolByName[AIDObject.Name];
            If LocalIDObject = Nil
               Then Begin
                  fSymbols.Add(AIDObject);
                  Result:= True;
               End;
         End;
      //----------------------------------------------------
      If AIDObject.InheritsFrom(TShape)
         Then Begin
            LocalIDObject:= ShapeByName[AIDObject.Name];
            If LocalIDObject = Nil
               Then Begin
                  fShapes.Add(AIDObject);
                  Result:= True;
               End;
         End;
   End;
End;
//********************************************************************
Procedure TGraphManager.Initialize;
Begin
   glClearDepth(1.0);				       // Depth Buffer Setup
   glEnable(GL_DEPTH_TEST);		     // Enable Depth Testing
   glDepthFunc(GL_LEQUAL);			   // The Type Of Depth Test To Do
   glShadeModel(GL_SMOOTH);        // Smoothe shading
   glEnable(GL_Texture_2D);        // Enable Textures
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
   //glEnable(GL_POINT_SMOOTH);      // Antaliasing
   glEnable(GL_LINE_SMOOTH);       // Antaliasing
   //glEnable(GL_POLYGON_SMOOTH);    // Antaliasing
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glEnable(GL_BLEND);             // Enable Blending

   With OGLBGColor Do
      glClearColor(Red/255,Green/255,Blue/255,Alpha/255);
End;
//********************************************************************
Procedure TGraphManager.Resize(AWidth,AHeight: Integer);
Begin
   glViewport(0,0,AWidth,AHeight);
End;
//********************************************************************
Procedure TGraphManager.PaintBegin;
Begin
   glClear(GL_COLOR_BUFFER_BIT OR GL_DEPTH_BUFFER_BIT);
End;
//********************************************************************
Procedure TGraphManager.PaintEnd;
Begin
   WindowSwapBuffers(fWindow);
End;
//********************************************************************
// No comment
Procedure TGraphManager.SetLineWidth(ALineWidth: Integer);
Begin
   glLineWidth(ALineWidth);
End;
//********************************************************************
Function TGraphManager.GetColorByIndex(AIndex: Integer): TRGBAColor;
Begin
   Result:= Nil;
   If ValidIndex(AIndex,fColors)
      Then Result:= fColors.List[AIndex];
End;
//********************************************************************
Function TGraphManager.GetFontByIndex(AIndex: Integer): TFreeType2Font;
Begin
   Result:= Nil;
   If ValidIndex(AIndex,fFonts)
      Then Result:= fFonts.List[AIndex];
End;
//********************************************************************
Function TGraphManager.GetImageByIndex(AIndex: Integer): TRGBAImage;
Begin
   Result:= Nil;
   If ValidIndex(AIndex,fImages)
      Then Result:= fImages.List[AIndex];
End;
//********************************************************************
Function TGraphManager.GetShapeByIndex(AIndex: Integer): TShape;
Begin
   Result:= Nil;
   If ValidIndex(AIndex,fShapes)
      Then Result:= fShapes.List[AIndex];
End;
//********************************************************************
Function TGraphManager.GetSymbolByIndex(AIndex: Integer): TSymbol;
Begin
   Result:= Nil;
   If ValidIndex(AIndex,fSymbols)
      Then Result:= fSymbols.List[AIndex];
End;
//********************************************************************
Function TGraphManager.GetColorByName(AName: String): TRGBAColor;
Begin
   Result:= TRGBAColor(GetItemByName(AName,fColors));
End;
//********************************************************************
Function TGraphManager.GetFontByName(AName: String): TFreeType2Font;
Begin
   Result:= TFreeType2Font(GetItemByName(AName,fFonts));
End;
//********************************************************************
Function TGraphManager.GetImageByName(AName: String): TRGBAImage;
Begin
   Result:= TRGBAImage(GetItemByName(AName,fImages));
End;
//********************************************************************
Function TGraphManager.GetShapeByName(AName: String): TShape;
Begin
   Result:= TShape(GetItemByName(AName,fShapes));
End;
//********************************************************************
Function TGraphManager.GetSymbolByName(AName: String): TSymbol;
Begin
   Result:= TSymbol(GetItemByName(AName,fSymbols));
End;
//********************************************************************
Function TGraphManager.GetItemByName(AName: String;
                      AList: TObjectList): TIDObject;
Var LocalObject: TIDObject; I: Integer; Found: Boolean;
Begin
   Result:= Nil;

   AName:= UpperCase(AName);
   I:=0;
   Found:= False;
   With AList Do While (I < Count) And (Not Found) Do Begin
      LocalObject:= List[I];
      If LocalObject.Name <> AName
         Then Inc(I) //================>
         Else Begin
            Result:= LocalObject;
            Found:= True;
         End;
   End;
End;
//********************************************************************
Function TGraphManager.ValidIndex(AIndex: Integer;
                   AList: TObjectList): Boolean;
Begin
   Result:= True;
   If (AIndex >= 0) And (AIndex < AList.Count)
      Then Result:= True;
End;
//********************************************************************
//********************************************************************
//   T I D O B J E C T
//********************************************************************
Constructor TIDObject.Create(AName: String);
Begin
   Inherited Create;
   SetName(AName);

End;
//********************************************************************
Procedure TIDObject.SetName(AName: String);
Begin
   fName:= UpperCase(AName);
End;
//********************************************************************
//********************************************************************
//   T I D O W N E D O B J E C T
//********************************************************************
Constructor TIDOwnedObject.Create(AName: String; AOwner: TCustomOwner);
Begin
   Inherited Create(AName);
   If Assigned (AOwner)
      Then fRegistered:= AOwner.Register(Self)
      Else fRegistered:= FALSE;
End;
//********************************************************************
//   T R G B A C O L O R
//********************************************************************
Constructor TRGBAColor.Create(AName: String; AOwner: TCustomOwner;
                     ARGBAColor: RGBAColor);
Begin
   Inherited Create(AName, AOwner);
   SetBaseColor(ARGBAColor);
End;
//********************************************************************
Procedure TRGBAColor.SetBaseColor(ARGBAColor: RGBAColor);
Begin
   fBaseColor:= ARGBAColor;
   Reset;
End;
//********************************************************************
Function TRGBAColor.GetColor: LongWord;
Begin
   Result:= RGBAToCardinal(fBaseColor);
End;
//********************************************************************
Procedure TRGBAColor.SetColor(AColor: LongWord);
Begin
   fBaseColor:= CardinalToRGBA(AColor);
   SetBaseColor(fBaseColor);
End;
//********************************************************************
Procedure TRGBAColor.Reset;
Begin
   fShownColor:= fBaseColor;
End;
//********************************************************************
// Set the color-profile in the range of [0..1] to create a color
// gradient: Black(0) -> BaseColor(0.5) -> White(1)
Procedure TRGBAColor.SetGradient(AGradient: Single);
//---------------------------------------------------------
Function GradeDown(AColor: Byte): Byte;
Begin
   Result:= Trunc(AColor*(1+AGradient));
End;
//---------------------------------------------------------
Function GradeUp(AColor: Byte): Byte;
Begin
   Result:= Trunc(AColor*(1-AGradient) + AGradient*255);
End;
//---------------------------------------------------------
Begin
   If (AGradient >= 0) And (AGradient <= 1)
      Then With fBaseColor Do Begin
         AGradient:= (AGradient - 0.5)*2;
         If AGradient < 0
            Then Begin // Range [-1..0]
               fShownColor.Red   := GradeDown(Red);
               fShownColor.Green := GradeDown(Green);
               fShownColor.Blue  := GradeDown(Blue);
            End
            Else Begin // Range [0..+1]
               fShownColor.Red   := GradeUp(Red);
               fShownColor.Green := GradeUp(Green);
               fShownColor.Blue  := GradeUp(Blue);
            End;
      End;
End;
//********************************************************************
Procedure TRGBAColor.Show;
Begin
   With fShownColor Do glColor4ub(Red,Green,Blue,Alpha);
End;
//********************************************************************
//********************************************************************
//   T R G B A I M A G E
//********************************************************************
Constructor TRGBAImage.Create(AName: String; AOwner: TCustomOwner);
Begin
   Inherited Create(AName,AOwner);
   fWidth:=0;
   fHeight:=0;
   fRGBAPixelMap:= Nil;
   fTexture:= 0;
   fAlphaColor:= SOLIDBLACK;
   fFlip:= FALSE;
   fFileFormat:= ifUnknown;
   fFileName:= 'Textures/' + Name;

   BuildImage;
End;
//********************************************************************
Destructor TRGBAImage.Destroy;
Begin
   If fTexture > 0
      Then glDeleteTextures(1, @fTexture);

   If fRGBAPixelMap <> Nil
      Then FreeMem(fRGBAPixelMap,fWidth*fHeight*4);
   Inherited Destroy;
End;
//********************************************************************
Function TRGBAImage.GetColor: LongWord;
Begin
   Result:= RGBAToCardinal(fAlphaColor);
End;
//********************************************************************
Procedure TRGBAImage.SetColor(AColor: LongWord);
Begin
   fAlphaColor:= CardinalToRGBA(AColor);
End;
//********************************************************************
Function TRGBAImage.BuildImage: Boolean;
Begin
   Result:= False;

   If FileExists(fFileName+'.BMP') Then fFileFormat:= ifBMP24Bit;
   If FileExists(fFileName+'.TGA') Then fFileFormat:= ifTGA32Bit;

   Case fFileFormat Of
   // ------------------------------------------
      ifBMP24Bit:
         If ImportBMP_BGR Then Begin
            Expand24To32Bit(fAlphaColor);
            //RGBToGrayScale;
            FlipHoriz;
            GenerateTexture;
            Result:= True;
         End;
   // ------------------------------------------
      ifTGA32Bit:
         If ImportTGA_BGRA Then Begin
            GenerateTexture;
            Result:= True;
         End;
   // ------------------------------------------
   End; // Of Case
End;
//********************************************************************
Procedure TRGBAImage.RGBToGrayScale;
Var
   RGBAPixel: ^TRGBAPixel;
   CountPixel,I: Integer; GrayByte: Byte;
Begin
   CountPixel:= fWidth*fHeight;
   RGBAPixel := fRGBAPixelMap;
   For I:= 0 To CountPixel-1 Do With RGBAPixel^ Do Begin
      GrayByte:= Round(0.299*Red + 0.587*Green + 0.114*Blue);
      Red   := GrayByte;
      Green := GrayByte;
      Blue  := GrayByte;
      Inc(RGBAPixel);
   End;
End;
//********************************************************************
Procedure TRGBAImage.Show(Visible: Boolean);
Begin
   If Visible
      Then glBindTexture(GL_TEXTURE_2D, fTexture)
      Else glBindTexture(GL_TEXTURE_2D, 0);
End;
//********************************************************************
Procedure TRGBAImage.FlipHoriz;
Var
   RGBAPixel1,RGBAPixel2: PRGBAPixel;
   TmpPixel: TRGBAPixel;
   Col,Row: Integer;
Begin
   For Row:= 0 To ((fHeight-1) Div 2) Do
      For Col:= 0 To fWidth-1 Do Begin

         RGBAPixel1:= fRGBAPixelMap;
         Inc(RGBAPixel1, fHeight*Row + Col);
         With TmpPixel Do Begin
            Red   := RGBAPixel1^.Red;
            Green := RGBAPixel1^.Green;
            Blue  := RGBAPixel1^.Blue;
            Alpha := RGBAPixel1^.Alpha;
         End;
         RGBAPixel2:= fRGBAPixelMap;
         Inc(RGBAPixel2,fHeight*(fHeight-Row-1)+Col);

         With RGBAPixel1^ Do Begin
            Red   := RGBAPixel2^.Red;
            Green := RGBAPixel2^.Green;
            Blue  := RGBAPixel2^.Blue;
            Alpha := RGBAPixel2^.Alpha;
         End;
         With RGBAPixel2^ Do Begin
            Red   := TmpPixel.Red;
            Green := TmpPixel.Green;
            Blue  := TmpPixel.Blue;
            Alpha := TmpPixel.Alpha;
         End;
      End;
End;
//********************************************************************
Procedure TRGBAImage.Expand24To32Bit(AlphaColor: RGBAColor);
Var
   RGBPixel: ^TBGRPixel;
   RGBAPixel: ^TRGBAPixel;
   CountPixel,I: Integer;
Begin
   CountPixel:= fWidth*fHeight;
   RGBPixel := fRGBAPixelMap; Inc(RGBPixel,CountPixel-1);
   RGBAPixel := fRGBAPixelMap; Inc(RGBAPixel,CountPixel-1);
   For I:= 0 To CountPixel-1 Do With RGBAPixel^ Do Begin

      Red   := RGBPixel^.Red;
      Green := RGBPixel^.Green;
      Blue  := RGBPixel^.Blue;

      If (Red   = AlphaColor.Red) And
         (Green = AlphaColor.Green) And
         (Blue  = AlphaColor.Blue)
         Then Alpha:= 0
         Else Alpha:= 255;

      Dec(RGBPixel);
      Dec(RGBAPixel);
   End;
End;
//********************************************************************
Procedure TRGBAImage.GenerateTexture;
Begin
   glGenTextures(1, @fTexture);
   glBindTexture(GL_TEXTURE_2D, fTexture);

   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,fWidth,fHeight,
      0,GL_RGBA,GL_UNSIGNED_BYTE,fRGBAPixelMap);

   glBindTexture(GL_TEXTURE_2D, 0);

   FreeMem(fRGBAPixelMap,fWidth*fHeight*4);
   fRGBAPixelMap:= Nil;
End;
//********************************************************************
Function TRGBAImage.ImportBMP_BGR: Boolean;
Begin
   fRGBAPixelMap:= LoadBMP_BGR(fFileName,fWidth,fHeight);

   If fRGBAPixelMap <> Nil
      Then Result:= True
      Else Result:= False;
End;
//********************************************************************
Function TRGBAImage.ImportTGA_BGRA: Boolean;
Begin
   fRGBAPixelMap:= LoadTGA_BGRA(fFileName,fWidth,fHeight);

   If fRGBAPixelMap <> Nil
      Then Result:= True
      Else Result:= False;
End;
//********************************************************************
//********************************************************************
//   T F R E E T Y P E 2 F O N T
//********************************************************************
Constructor TFreeType2Font.Create(AName: String;
                                 AOwner: TCustomOwner);
Begin
   Inherited Create(AName,AOwner);
   fFileName:= 'Fonts/' + Name +'.TTF';;
   fColor:= SOLIDBLACK;
   fSize:= 24;
   fStyle:= 0;
   fMargin:= 2;

   fBitmap16bpp:= Nil;
   BuildFont;
End;
//********************************************************************
Destructor TFreeType2Font.Destroy;
Begin
   If fLetterBase > 0 Then Begin
      glDeleteLists(fLetterBase,256);
      glDeleteTextures(1,@fTexture);
   End;
   Inherited Destroy;
End;
//********************************************************************
Function TFreeType2Font.BuildFont: Boolean;
Var Error,FaceIndex: Integer;
Begin
   Result:= False;

   Error:= FT_Init_FreeType(fFreeTypeLib);
   If Error <> 0
      Then; // ERROR

   FaceIndex:= 0;
   Repeat
      Error:= FT_New_Face(fFreeTypeLib,PChar(fFileName),
                          FaceIndex,fGlyphFace);
      If Error <> 0
         Then; // ERROR

      Inc(FaceIndex);
   Until (fStyle = fGlyphFace.Style_Flags) Or
         (FaceIndex >= fGlyphFace.Num_Faces);

   If fStyle <> fGlyphFace.Style_Flags
      Then; // ERROR

   //-------------------------
   GetFontMetrics;
   If GenerateBitmap Then Begin
      DrawLettersOnBitmap;
      GenerateTexture;
      BuildDisplayLists;
      Result:= True;
   End;
   //-------------------------

   Error:= FT_Done_Face(fGlyphFace);
   If Error <> 0
      Then; //ERROR

   Error:= FT_Done_FreeType(fFreeTypeLib);
   If Error <> 0
      Then; //ERROR
End;
//********************************************************************
Procedure TFreeType2Font.BuildDisplayLists;
Var TxLeft,TxRight,TxBottom,TxTop,
    VxLeft,VxRight,VxBottom,VxTop: Single;
    I: Integer;
Begin
   If fLetterBase > 0
      Then glDeleteLists(fLetterBase,256);
   fLetterBase:= glGenLists(256);

   For I:= 0 To 255 Do Begin
      TxLeft:= ((I Mod 16)*fLetterWidthMax)/(fTextureWidth);
      TxRight:= TxLeft + fLetterWidth[I]/(fTextureWidth);

      //TxTop:= ((I Div 16)*fLetterHeightMax)/(fTextureHeight);
      TxTop:= (((I Div 16)*fLetterHeightMax)+fSize/8)/(fTextureHeight);
      TxBottom:= TxTop + (fLetterHeightMax)/(fTextureHeight);

      VxLeft   := 0;
      VxRight  := fLetterWidth[I];
      VxTop    := 0;
      VxBottom := fLetterHeightMax;

      glNewList(fLetterBase+I, GL_COMPILE);
      glBegin(GL_QUADS);
         glTexCoord2f(TxLeft, TxBottom); glVertex2f(VxLeft, VxBottom);
         glTexCoord2f(TxRight,TxBottom); glVertex2f(VxRight,VxBottom);
         glTexCoord2f(TxRight,TxTop);    glVertex2f(VxRight,VxTop);
         glTexCoord2f(TxLeft, TxTop);    glVertex2f(VxLeft, VxTop);
      glEnd;
      glTranslatef(fLetterWidth[I]+1,0,0);
      glEndList;
   End;
End;
//********************************************************************
Procedure TFreeType2Font.Copy8bppTo16bpp
   (Source: PMonoPixel; Dest: PLumenPixel;
    SourceX,SourceY,SourceWidth,SourceHeight,
    DestX,DestY,DestWidth,DestHeight: Integer);
//--------------------------------------------------------------------
Function Get8bppPixel(ABitmap: PMonoPixel;
                 X,Y,Width,Height: Integer): TMonoPixel;
Var LMonoPixel: PMonoPixel;
Begin
   LMonoPixel:= ABitmap;
   Inc(LMonoPixel,Width*Y + X); // Linewise advancement
   Result:= LMonoPixel^;
End;
//--------------------------------------------------------------------
Procedure Set16bppPixel(ABitmap: PLumenPixel; APixel: TLumenPixel;
                 X,Y,Width,Height: Integer);
Var LLumenPixel: PLumenPixel;
Begin
   LLumenPixel:= ABitmap;
   Inc(LLumenPixel,Width*Y + X); // Linewise advancement
   With LLumenPixel^ Do Begin
      Lumen:= APixel.Lumen;
      Alpha:= APixel.Alpha;
   End;
End;
//--------------------------------------------------------------------
Var
   MonoPixel: TMonoPixel;
   LumenPixel: TLumenPixel;
   Col,Row: Integer;
Begin
   For Row:= 0 To SourceHeight-1 Do
      For Col:= 0 To SourceWidth-1 Do
      Begin
         MonoPixel:= Get8bppPixel
            (Source,SourceX+Col,SourceY+Row,SourceWidth,SourceHeight);
         With LumenPixel Do Begin
            Lumen:= 255;
            Alpha:= MonoPixel;
         End;
         Set16bppPixel(Dest,LumenPixel,DestX+Col,DestY+Row,
            DestWidth,DestHeight);
      End;
End;
//********************************************************************
Procedure TFreeType2Font.GetFontMetrics;
Var
   PGlyph: PFT_Glyph;
   GlyphBitmap: TFT_BitmapGlyph;
   PGlyphBitmap: PFT_BitmapGlyph;
   Error,I,MaxAscend,MaxDescend,GlyphIndex: Integer;
Begin
   Error:= FT_Set_Char_Size(fGlyphFace,0,fSize*64,60,60);
   If Error <> 0
      Then; // ERROR

   fLetterWidthMax := 0;
   fLetterHeightMax:= 0;
   MaxAscend:= 0;
   MaxDescend:= 0;

   For I:= 0 To 255 Do Begin
      // Load glyph
      GlyphIndex:= FT_Get_Char_Index(fGlyphFace,I);
      FT_Load_Glyph(fGlyphFace,GlyphIndex,FT_LOAD_DEFAULT);
      FT_Render_Glyph(fGlyphFace.Glyph,FT_RENDER_MODE_NORMAL);
      With fGlyphFace.Glyph^ Do Begin
         MaxAscend:= Max(Bitmap_Top,MaxAscend);
         MaxDescend:= Max(Bitmap.Rows-Bitmap_Top, MaxDescend);
      End;
      fLetterWidth[I]:= fGlyphFace.Glyph.Advance.X/64;
      fLetterWidthMax := Max(fLetterWidthMax,Trunc(fLetterWidth[I]));
   End;
   fLetterHeightMax:= MaxAscend+MaxDescend+fMargin;
End;
//********************************************************************
Function TFreeType2Font.GenerateBitmap: Boolean;
Var I: Integer;
Begin
   Result:= False;

   // Enhance font bitmap to Power-Of-Two
   fTextureWidth := (fLetterWidthMax)*16;
   fTextureHeight:= (fLetterHeightMax)*16;

   I:=0;
   Repeat Inc(I);
   Until fTextureWidth <= Power(2,I);
   fTextureWidth:= Trunc(Power(2,I));

   I:=0;
   Repeat Inc(I);
   Until fTextureHeight <= Power(2,I);
   fTextureHeight:= Trunc(Power(2,I));

   // Check if texture size too big
   glGetIntegerv(GL_MAX_TEXTURE_SIZE,@I);
   If Max(fTextureWidth,fTextureHeight) <= I
      Then Begin
         // Create empty 16-bit-bitmap for OpenGL
         I:= fTextureWidth*fTextureHeight*2;
         GetMem(fBitmap16bpp,I);
         FillChar(fBitmap16bpp^,I,#0);

         Result:= True;
      End;
End;
//********************************************************************
Procedure TFreeType2Font.DrawLettersOnBitmap;
Var
   PGlyph: PFT_Glyph;
   GlyphBitmap: TFT_BitmapGlyph;
   PGlyphBitmap: PFT_BitmapGlyph;
   LetterBitmap8bpp: PMonoPixel;
   GlyphIndex,YDist,Error,LCol,LRow,I: Integer;
   LImage: TRGBAImage;
Begin
   For I:= 0 To 255 Do Begin
      // Load glyph
      Error:= FT_Load_Char(fGlyphFace,I,FT_LOAD_DEFAULT);
      If Error <> 0
         Then; // ERROR

      // Extract glyph image and get 8bpp bitmap
      PGlyph:= Nil;
      Error:= FT_Get_Glyph(fGlyphFace.Glyph, PGlyph);
      If Error <> 0
         Then; // ERROR

      Error:= FT_Glyph_To_Bitmap(PGlyph,FT_RENDER_MODE_NORMAL,Nil,True);
      If Error <> 0
         Then; // ERROR

      PGlyphBitmap:= PFT_BitmapGlyph(PGlyph);
      GlyphBitmap:= PGlyphBitmap^;
      LetterBitmap8bpp:= PGlyphBitmap.Bitmap.Buffer;

      // Copy letter bitmap to font bitmap
      LCol:= (I Mod 16) * fLetterWidthMax;
      LRow:= (I Div 16) * fLetterHeightMax;
      YDist:= fLetterHeightMax-GlyphBitmap.Top-2;
      If YDist < 0 Then YDist:= 0;
      With GlyphBitmap.Bitmap Do Copy8bppTo16bpp
         (LetterBitmap8bpp,fBitmap16bpp,
          0,0,Width,Rows,
          LCol + GlyphBitmap.Left,
          LRow + YDist,
          fTextureWidth,fTextureHeight);

      // Release letter bitmap
      FT_Done_Glyph(PGlyph);
   End;
End;
//********************************************************************
Procedure TFreeType2Font.GenerateTexture;
Begin
   If fTexture > 0
      Then glDeleteTextures(1,@fTexture);

   glGenTextures(1, @fTexture);

   glBindTexture(GL_TEXTURE_2D, fTexture);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,fTextureWidth,fTextureHeight,
         0,GL_LUMINANCE_ALPHA,GL_UNSIGNED_BYTE,fBitmap16bpp);

   glBindTexture(GL_TEXTURE_2D, 0);
   FreeMem(fBitmap16bpp);
   fBitmap16bpp:= Nil;
End;
//********************************************************************
Procedure TFreeType2Font.Show(XPos,YPos: Integer; AString: String);
Var
   Viewport: Array [0..3] Of Integer;
Begin
   glMatrixMode(GL_PROJECTION);
   glPushMatrix;
   glLoadIdentity;
   glGetIntegerv(GL_VIEWPORT, @Viewport);
   glOrtho(0,ViewPort[2],ViewPort[3],0,+1,-1);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix;
   glLoadIdentity;

   glTranslatef(XPos,YPos,0);
   With fColor Do glColor4ub(Red,Green,Blue,Alpha);
   glBindTexture(GL_TEXTURE_2D, fTexture);

   glListBase(fLetterBase);
   glCallLists(Length(AString),GL_UNSIGNED_BYTE,PChar(AString));

   glBindTexture(GL_TEXTURE_2D,0);

   glMatrixMode(GL_MODELVIEW);
   glPopMatrix;
   glMatrixMode(GL_PROJECTION);
   glPopMatrix;
End;
//********************************************************************
Function TFreeType2Font.TextHeight(AString: String): Word;
Begin
   Result:= fLetterHeightMax;
End;
//********************************************************************
Function TFreeType2Font.TextWidth(AString: String): LongWord;
Var LocalWidth: Double; I,Len: Integer; Letter: Char;
Begin
   LocalWidth:= 0;
   Len:= Length(AString);
   For I:= 1 To Len Do Begin
      Letter:= AString[I];
      LocalWidth:= LocalWidth+fLetterWidth[Ord(Letter)];
   End;
   Result:= Trunc(LocalWidth);
End;
//********************************************************************
Function TFreeType2Font.GetColor: LongWord;
Begin
   Result:= RGBAToCardinal(fColor);
End;
//********************************************************************
Procedure TFreeType2Font.SetColor(AColor: LongWord);
Begin
   fColor:= CardinalToRGBA(AColor);
End;
//********************************************************************
Procedure TFreeType2Font.SetSize(ASize: Integer);
Begin
   If (ASize > 8) And (ASize <= 36)
      Then Begin
         fSize:= ASize;
         BuildFont;
      End;
End;
//********************************************************************
//********************************************************************
//   T R E N D E R E R
//********************************************************************
Function TRenderer.EllipsePoint(Angle: Integer): TVertex;
Var
   DegToRad,Sinus,Cosinus: Single;
Begin
   DegToRad:= Angle*PiDiv180;
   With Result Do Begin
      PtX:= Cos(DegToRad) * 0.5;
      PtY:= Sin(DegToRad) * 0.5;
   End;
End;
//********************************************************************
Procedure TRenderer.OrthoSandBoxOFF;
Begin
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix;
   glMatrixMode(GL_PROJECTION);
   glPopMatrix;
End;
//********************************************************************
Procedure TRenderer.OrthoSandBoxON;
Var
   Viewport: Array [0..3] Of Integer;
Begin
   glMatrixMode(GL_PROJECTION);
   glPushMatrix;
   glLoadIdentity;
   glGetIntegerv(GL_VIEWPORT, @Viewport);
   glOrtho(0,ViewPort[2],ViewPort[3],0,+1,-1);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix;
   glLoadIdentity;
End;
//********************************************************************
Procedure TRenderer.Transform(ALeft,ATop,AWidth,AHeight: Integer);
Begin
   glTranslatef(ALeft+AWidth/2,ATop+AHeight/2,0);
   glScalef(AWidth,AHeight,1);
End;
//********************************************************************
//   T S Y M B O L
//********************************************************************
Constructor TSymbol.Create(AName: String; AOwner: TCustomOwner);
Begin
   Inherited Create(AName,AOwner);
   Build;
End;
//********************************************************************
Destructor TSymbol.Destroy;
Begin
   If fSymbol > 0
      Then glDeleteLists(fSymbol,1);
   Inherited Destroy;
End;
//********************************************************************
Procedure TSymbol.Show(ALeft,ATop,AWidth,AHeight: Integer;
         AColor: RGBAColor; ASpecification: Integer);
Begin
   OrthoSandBoxON;

   Transform(ALeft,ATop,AWidth,AHeight);
   glLineWidth(2);
   With AColor Do glColor4ub(Red,Green,Blue,Alpha);
   glCallList(fSymbol);

   OrthoSandBoxOFF;
End;
//********************************************************************
Procedure TArrow.Build;
Var
   Cos00,Sin00,Cos60,Sin60: Single;
Begin
   Cos00:= 1;
   Sin00:= 0;
   Cos60:= Cos(60*Pi/180);
   Sin60:= Sin(60*Pi/180);

   If fSymbol > 0
      Then glDeleteLists(fSymbol,1);
   fSymbol:= glGenLists(1);

   glNewList(fSymbol, GL_COMPILE);
      glBegin(GL_TRIANGLES);
         glVertex2f(+0.25*Cos00, +0.25*Sin00);
         glVertex2f(-0.25*Cos60, +0.25*Sin60);
         glVertex2f(-0.25*Cos60, -0.25*Sin60);
      glEnd;
   glEndList;
End;
//********************************************************************
Procedure TArrow.Show(ALeft,ATop,AWidth,AHeight: Integer;
         AColor: RGBAColor; ASpecification: Integer);
Begin
   OrthoSandBoxON;

   Transform(ALeft,ATop,AWidth,AHeight);
   Case ASpecification Of
      1: glRotatef(-90,0,0,1);   // UP
      2: glRotatef(+180,0,0,1);  // LEFT
      3: glRotatef(+90,0,0,1);   // DOWN
      Else ;                     // RIGHT + all other
   End;

   With AColor Do glColor4ub(Red,Green,Blue,Alpha);
   glCallList(fSymbol);

   OrthoSandBoxOFF;
End;
//********************************************************************
Procedure TCross.Build;
Begin
   If fSymbol > 0
      Then glDeleteLists(fSymbol,1);
   fSymbol:= glGenLists(1);

   glNewList(fSymbol, GL_COMPILE);
      glBegin(GL_LINES);
         glVertex2f(-0.25,+0.25);  // Upper left
         glVertex2f(+0.25,-0.25);  // Lower right
         glVertex2f(+0.25,+0.25);  // Upper right
         glVertex2f(-0.25,-0.25);  // Lower left
      glEnd;
   glEndList;
End;
//********************************************************************
// Hook has to be build upside down
Procedure THook.Build;
Begin
   If fSymbol > 0
      Then glDeleteLists(fSymbol,1);
   fSymbol:= glGenLists(1);

   glNewList(fSymbol, GL_COMPILE);
      glBegin(GL_LINES);
         glVertex2f(-0.15,+0.00);
         glVertex2f(+0.00,+0.25);
         glVertex2f(+0.00,+0.25);
         glVertex2f(+0.30,-0.30);
      glEnd;
   glEndList;
End;
//********************************************************************
Procedure TSpot.Build;
Var Step,Angle: Integer;
Begin
   If fSymbol > 0
      Then glDeleteLists(fSymbol,1);
   fSymbol:= glGenLists(1);

   Step:= 15;
   Angle:= 0;
   glNewList(fSymbol, GL_COMPILE);
      glBegin(GL_POLYGON);
         While Angle < 360 Do Begin
            With EllipsePoint(Angle) Do glVertex2f(PtX,PtY);
            Inc(Angle,Step); //===============>
         End;
      glEnd;
   glEndList;
End;
//********************************************************************
//   T S H A P E
//********************************************************************
Constructor TShape.Create(AName: String; AOwner: TCustomOwner);
Begin
   Inherited Create(AName,AOwner);
   fSilhouette:= TList.Create;
   fTriangles:= TList.Create;
   fFrame:= TList.Create;
   fFileName:= 'Silhouettes/'+ Name;

   If Not Load
      Then GenSilhouette;
   Triangulate;
   BuildShape;
   BuildFrame;
End;
//********************************************************************
Destructor TShape.Destroy;
Begin
   ClearLists;
   fFrame.Free;
   fTriangles.Free;
   fSilhouette.Free;

   Inherited Destroy;
End;
//********************************************************************
Procedure TShape.ClearLists;
Var LocalVertex: PTexVertex; I: Integer;
Begin
   If fShape > 0
      Then glDeleteLists(fShape,1);

   ClearFrame; // Deletes all frame-lists

   fTriangles.Clear; // Contains only pointers of fSilhouette

   With fSilhouette Do If Count > 0
      Then For I:= 0 To Count-1 Do Begin
         LocalVertex:= Items[I];
         Dispose(LocalVertex);
      End;
   fSilhouette.Clear;
End;
//********************************************************************
Procedure TShape.ClearFrame;
Var I: Integer; LocalFrame: PWord;
Begin
   With fFrame Do If Count > 0
      Then For I:= 0 To Count-1 Do Begin
         LocalFrame:= Items[I];
         If LocalFrame^ > 0
            Then glDeleteLists(LocalFrame^,1);
         Dispose(LocalFrame);
      End;
   fFrame.Clear;
End;
//********************************************************************
Procedure TShape.AddVertex(ATexX,ATexY,AX,AY: Single);
Var LocalVertex: PTexVertex;
Begin
   New(LocalVertex);
   With LocalVertex^ Do Begin
      TexX:= ATexX;
      TexY:= ATexY;
      PtX:= AX;
      PtY:= AY;
   End;
   fSilhouette.Add(LocalVertex);
End;
//********************************************************************
Procedure TShape.Show(ALeft,ATop,AWidth,AHeight: Integer;
    ADrawMode: TDrawMode; AColor: TRGBAColor; AImage: TRGBAImage;
    AFrameWidth: Integer; AFrameState: TFrameState);
Var
   LocalFrame: PWord; MaxVal,MaxRange: Single;
   I,J: Integer; LColor: Byte;
//--------------------------------------------------------
Function ColorLight(CountVal,MaxVal,MaxRange,Value: Single): Byte;
Var Factor0,Factor1,Factor2: Single; Interim: Byte;
Begin
   Factor0:= (CountVal-1)/2;
   Factor1:= Factor0*Factor0;
   Factor2:= (Value-Factor0)*(Value-Factor0);
   Interim:= Trunc((Factor1-Factor2)/MaxVal*MaxRange);
   If MaxRange < 255
      Then Result:= Trunc(Interim+(255-MaxRange));
End;
//--------------------------------------------------------
Function InversColor(AColor: Byte): Byte;
Begin
   Result:= -AColor+255;
End;
//--------------------------------------------------------
Begin
   OrthoSandBoxON;

   AColor.Show;
   With AImage Do If Texture <> 0
      Then Show(TRUE);

   Transform(ALeft,ATop,AWidth,AHeight);

   If (ADrawMode = dmAll) Or (ADrawMode = dmShapeOnly)
      Then glCallList(fShape);

   With AImage Do If Texture <> 0
      Then Show(FALSE);

   If (ADrawMode = dmAll) Or (ADrawMode = dmFrameOnly)
      Then Begin
         glLineWidth(1.5);
         With fFrame Do If Count > 0 Then Begin
            For J:= 0 To AFrameWidth-1 Do Begin
               MaxRange:= 255/(J+1);
               MaxVal:= (Count/2)*(Count/2);
               glLoadIdentity;
               Transform(ALeft+J,ATop+J,AWidth-2*J,AHeight-2*J);
               For I:= 0 To Count-1 Do Begin
                  If AFrameState = fsFlat
                     Then glColor4ub(128,128,128,255)
                     Else Begin
                        LColor:= ColorLight(Count,MaxVal,MaxRange,I);
                        If AFrameState = fsLowered
                           Then LColor:= InversColor(LColor);
                        glColor4ub(LColor,LColor,LColor,255);
                     End;
                  LocalFrame:= Items[I];
                  glCallList(LocalFrame^);
               End;
            End;
         End;
      End;

   OrthoSandBoxOFF;
End;
//********************************************************************
Procedure TShape.BuildShape;
Var
   LocalVertex: PTexVertex; I: Integer;
Begin
   If fShape > 0
      Then glDeleteLists(fShape,1);
   fShape:= glGenLists(1);

   glNewList(fShape, GL_COMPILE);
      glBegin(GL_POLYGON);
         With fSilhouette Do For I:= 0 To Count-1 Do Begin
            LocalVertex:= Items[I];
            With LocalVertex^ Do Begin
               glTexCoord2f(TexX,TexY);
               glVertex2f(PtX,PtY);
            End;
         End;
      glEnd;
   glEndList;

   ErrorString:= gluErrorString(glGetError);
End;
//********************************************************************
// Frame consists of frame-pieces
Procedure TShape.BuildFrame;
Var
   Vertex1,Vertex2: PTexVertex; I: Integer; LocalFrame: PWord;
Begin
   If fFrame.Count > 0
      Then ClearFrame;

   With fSilhouette Do Begin
      Vertex1:= Items[Count-1];
      For I:= 0 To Count-1 Do Begin
         New(LocalFrame);
         fFrame.Add(LocalFrame);
         LocalFrame^:= glGenLists(1);

         // Generates one piece of frame
         glNewList(LocalFrame^, GL_COMPILE);
            glBegin(GL_LINES);
               Vertex2:= Items[I];
               With Vertex1^ Do glVertex2f(PtX,PtY);
               With Vertex2^ Do glVertex2f(PtX,PtY);
               Vertex1:= Vertex2;
            glEnd;
         glEndList;
      End;
   End;
End;
//********************************************************************
Function TShape.Load: Boolean;
Begin
   Result:= False;
   If FileExists(fFileName)
      Then;
End;
//********************************************************************
Procedure TShape.Triangulate;
Begin
End;
//********************************************************************
// Important: Begin with DARKEST Point
Procedure TRectangle.GenSilhouette;
Begin
   AddVertex(0,1, -0.5,+0.5);  // Upper left
   AddVertex(0,0, -0.5,-0.5);  // Lower left
   AddVertex(1,0, +0.5,-0.5);  // Lower right
   AddVertex(1,1, +0.5,+0.5);  // Upper right
End;
//********************************************************************
// Important: Begin with DARKEST Point
Procedure TEllipse.GenSilhouette;
Var Angle,Darkness,Step: Integer;
Begin
   Step:=5;
   Darkness:=+45; // Upside down because 0/0 = upper left

   Angle:= Darkness;
   While Angle < (360+Darkness) Do Begin
      With EllipsePoint(Angle) Do
         AddVertex(0.5+PtX,0.5+PtY,PtX,PtY);
      Inc(Angle,Step); //===============>
   End;
End;
//********************************************************************
// Important: Begin with DARKEST Point
Procedure THexagon.GenSilhouette;
Var Angle,Darkness,Step: Integer;
Begin
   Step:=60;
   Darkness:=+90; // Upside down because 0/0 = upper left

   Angle:= Darkness;
   While Angle < (360+Darkness) Do Begin
      With EllipsePoint(Angle) Do
         AddVertex(0.5+PtX,0.5+PtY,PtX,PtY);
      Inc(Angle,Step); //===============>
   End;
End;
//********************************************************************
Constructor TRoundRect.Create(AName: String; AOwner: TCustomOwner);
Begin
   SetEllipseParams;

   Inherited Create(AName,AOwner);
End;
//********************************************************************
Procedure TRoundRect.SetEllipseParams;
Begin
   fEllipseX:= 0.2;
   fEllipseY:= fEllipseX;
End;
//********************************************************************
// Important: Begin with DARKEST Point
Procedure TRoundRect.GenSilhouette;
Var X,Y: Single; Angle,Step,Darkness: Integer;
Begin
   Step:=5;
   Darkness:= +45;

   Angle:=Darkness;
   While Angle < 90 Do Begin  // 1.Quadrant (halb)
      With EllipsePoint(Angle) Do Begin
         X:= +0.5 *(1-fEllipseX) + PtX * fEllipseX;
         Y:= +0.5 *(1-fEllipseY) + PtY * fEllipseY;
         AddVertex(0.5+X,0.5+Y,X,Y);
      End;
      Inc(Angle,Step); //===============>
   End;

   Angle:=90;
   While Angle < 180 Do Begin  // 2.Quadrant
      With EllipsePoint(Angle) Do Begin
         X:= -0.5 *(1-fEllipseX) + PtX * fEllipseX;
         Y:= +0.5 *(1-fEllipseY) + PtY * fEllipseY;
         AddVertex(0.5+X,0.5+Y,X,Y);
      End;
      Inc(Angle,Step); //===============>
   End;

   Angle:=180;
   While Angle < 270 Do Begin  // 3.Quadrant
      With EllipsePoint(Angle) Do Begin
         X:= -0.5 *(1-fEllipseX) + PtX * fEllipseX;
         Y:= -0.5 *(1-fEllipseY) + PtY * fEllipseY;
         AddVertex(0.5+X,0.5+Y,X,Y);
      End;
      Inc(Angle,Step); //===============>
   End;

   Angle:=270;
   While Angle < 360 Do Begin  // 4.Quadrant
      With EllipsePoint(Angle) Do Begin
         X:= +0.5 *(1-fEllipseX) + PtX * fEllipseX;
         Y:= -0.5 *(1-fEllipseY) + PtY * fEllipseY;
         AddVertex(0.5+X,0.5+Y,X,Y);
      End;
      Inc(Angle,Step); //===============>
   End;

   Angle:= 0;
   While Angle < Darkness Do Begin  // 1.Quadrant(halb)
      With EllipsePoint(Angle) Do Begin
         X:= +0.5 *(1-fEllipseX) + PtX * fEllipseX;
         Y:= +0.5 *(1-fEllipseY) + PtY * fEllipseY;
         AddVertex(0.5+X,0.5+Y,X,Y);
      End;
      Inc(Angle,Step); //===============>
   End;
End;
//********************************************************************
Procedure TRoundRectHoriz.SetEllipseParams;
Begin
   fEllipseX:= 0.1;
   fEllipseY:= 1;
End;
//********************************************************************
Procedure TRoundRectVert.SetEllipseParams;
Begin
   fEllipseX:= 1;
   fEllipseY:= 0.1;
End;
//********************************************************************
//********************************************************************
INITIALIZATION

OGLBGColor:= SOLIDBLACK;

End.

