//********************************************************************
UNIT ImageLoader;
//********************************************************************
INTERFACE

USES
   SysUtils,
   Classes;

TYPE
TImageFileFormat = (ifUnknown, ifBMP24Bit, ifTGA32Bit);

Function LoadBMP_BGR(AFileName: String;
                     Var AWidth, AHeight: Integer): Pointer;
Function LoadTGA_BGRA(AFileName: String;
            Var AWidth, AHeight: Integer): Pointer;
//********************************************************************
IMPLEMENTATION
//********************************************************************
TYPE
   TRGBAPixel = Packed Record Red,Green,Blue,Alpha: Byte; End;
   PRGBAPixel = ^TRGBAPixel;
//********************************************************************
   TBitmapFileHeader = Packed Record
      bfType: Word;
      bfSize: LongWord;
      bfReserved1: Word;
      bfReserved2: Word;
      bfOffBits: LongWord;
   End;
//********************************************************************
   TBitmapInfoHeader = Packed Record
      bmSize: LongWord;           // The size of this header (always sizeof(BITMAPINFOHEADER))
      bmWidth: Integer;           // The width of the bitmap
      bmHeight: Integer;          // The height of the bitmap
      bmPlanes: Word;             // The planes in the bitmap (usually 1)
      bmBitsPerPix: Word;         // The bits per pixel (8 for 256 color bitmaps)
      bmCompression: LongWord;    // Type of compression (I use no compression or BI_RGB)
      bmSizeImage: LongWord;      // Total memory used by the bitmap (width * height in a 256 color bitmap)
      bmXPelsPerMeter: Integer;   // Pixels per meter along the X-axis (0 if you don't care)
      bmYPelsPerMeter: Integer;   //   "     "    "     "    "  Y-axis (0 if you don't care)
      bmClrUsed: LongWord;        // Usually 0 (I think) or number of colors in bitmap's palette
      bmClrImportant: LongWord;   // Always 0 (All colors are important)
   End;
//********************************************************************
{All Targa formats are identified by a Data Type field, which is a one
byte binary integer located in byte three of the file.  The various
file types specified by this field are as follows:

       0  -  No image data included.
       1  -  Uncompressed, color-mapped images.
       2  -  Uncompressed, RGB images.
       3  -  Uncompressed, black and white images.
       9  -  Runlength encoded color-mapped images.
      10  -  Runlength encoded RGB images.
      11  -  Compressed, black and white images.
      32  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.
      33  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.  4-pass quadtree-type process.
}
   TTargaHeader = Packed Record  // For Unmapped RGB images (ImageType=2)
      ImageIdentLen: Byte;  // Length of the Image Identification Field (see below)
      ColorMapType: Byte;   // Color Map Type (0=NoColorMap,1=Colormap exists)
      ImageDataType: Byte;  // See description above
      ColMapSpec: Array[0..4] of Byte; // Ignored: Color Map Specification

      OrigX: Word;   // X coordinate of the lower left corner
      OrigY: Word;   // Y coordinate of the lower left corner
      tgWidth: Word;   // Width of the image in pixels
      tgHeight: Word;  // Height of the image in pixels
      tgBitsPerPix: Byte;  // Number of bits in a pixel: 16/24/32
      ImageDescr: Byte;  // Image Descriptor Byte
      // ImageIdent  // Image Identification Field, usually omitted
   End;
//********************************************************************
Function SwapRGBAPixel(ARGBAPixel: PRGBAPixel): Boolean;
Var LocalPixel: TRGBAPixel;
Begin
   With ARGBAPixel^ Do Begin
      LocalPixel.Red:= Blue;
      LocalPixel.Blue:= Red;

      Red:= LocalPixel.Red;
      Blue:= LocalPixel.Blue;
   End;
End;
//********************************************************************
Function LoadBMP_BGR(AFileName: String;
            Var AWidth, AHeight: Integer): Pointer;
Var
   FileHeader: TBitmapFileHeader;
   InfoHeader: TBitmapInfoHeader;
   LocalStream: TFileStream;
   LocalFileName: String;
   LocalPixelMap: Pointer;
   MemorySize: Longword;
   CountBytes: Longword;
Begin
   Result:= Nil;

   LocalFileName:= AFileName + '.BMP';

   If FileExists(LocalFileName) Then Begin
      LocalStream:= TFileStream.Create(LocalFileName,
                                fmOpenRead,fmShareDenyWrite);
      FillChar(FileHeader, SizeOf(FileHeader), #0);
      FillChar(InfoHeader, SizeOf(InfoHeader), #0);
      LocalStream.Read(FileHeader, SizeOf(FileHeader));
      LocalStream.Read(InfoHeader, SizeOf(InfoHeader));
      With InfoHeader Do If (bmBitsPerPix = 24) And (bmCompression = 0)
      Then Begin
         MemorySize:= bmWidth*bmHeight*4;
         GetMem(LocalPixelMap,MemorySize);
         CountBytes:= LocalStream.Read(LocalPixelMap^,bmWidth*bmHeight*3);
         If CountBytes = bmWidth*bmHeight*3
            Then Begin
               AWidth:= bmWidth;
               AHeight:= bmHeight;
               Result:= LocalPixelMap;
            End
            Else FreeMem(LocalPixelMap,MemorySize);
      End;
      LocalStream.Free;
   End;
End;
//********************************************************************
Function LoadTGA_Uncompressed(AFileStream: TFileStream;
           AImageSize: LongWord; APixelMap: Pointer): Boolean;
Var CountBytes: Longword;
Begin
   Result:= False;

   CountBytes:= AFileStream.Read(APixelMap^, AImageSize);
   If CountBytes = AImageSize
      Then Result:= True;
End;
//********************************************************************
Function LoadTGA_Compressed(AFileStream: TFileStream;
           AImageSize: LongWord; APixelMap: Pointer): Boolean;
Var
   CountPixel,ImagePixel: Longword; I: Integer; Header: Byte;
   ReadPixel: TRGBAPixel;
   WritePixel: PRGBAPixel;
Begin
   Result:= False;
   WritePixel:= APixelMap;
   ImagePixel:= AImageSize Div 4;
   CountPixel:= 0;

   With AFileStream Do
   While (Position < Size) And (CountPixel < ImagePixel) Do Begin
      Read(Header,1);
      If Header < 128
         Then Begin // Raw data: Copy the following pixels
            Inc(Header);
            Read(WritePixel^,Header*4);
            Inc(WritePixel,Header);
            Inc(CountPixel,Header);
         End
         Else Begin // Compressed data: Multiply the one following pixel
            Dec(Header,127);
            Read(ReadPixel,4);
            For I:= 0 To Header-1 Do With WritePixel^ Do Begin
               Red   := ReadPixel.Red;
               Green := ReadPixel.Green;
               Blue  := ReadPixel.Blue;
               Alpha := ReadPixel.Alpha;

               Inc(WritePixel);
            End;
            Inc(CountPixel,Header);
         End;
   End;

   If CountPixel = ImagePixel
      Then Result:= True;
End;
//********************************************************************
// Gimp: Ebene / Transparenz / Add Alphakchannel
Function LoadTGA_BGRA(AFileName: String;
            Var AWidth, AHeight: Integer): Pointer;
Var
   FileHeader: TTargaHeader;
   LocalStream: TFileStream;
   LocalFileName: String;
   LocalPixelMap: Pointer;
   ImageSize: LongWord;
   Loaded: Boolean;
   RGBAPixel: PRGBAPixel;
   CountPixel,I: LongWord;
Begin
   Result:= Nil;

   LocalFileName:= AFileName + '.TGA';

   If FileExists(LocalFileName) Then Begin
      LocalStream:= TFileStream.Create(LocalFileName,
                                fmOpenRead,fmShareDenyWrite);
      FillChar(FileHeader, SizeOf(FileHeader), #0);
      LocalStream.Read(FileHeader, SizeOf(FileHeader));
      With FileHeader Do
         If (tgBitsPerPix = 32) Then Begin
            ImageSize:= tgWidth*tgHeight*4;
            GetMem(LocalPixelMap,ImageSize);
            Case ImageDataType Of
                2: Loaded:= LoadTGA_Uncompressed
                       (LocalStream,ImageSize,LocalPixelmap);
               10: Loaded:= LoadTGA_Compressed
                       (LocalStream,ImageSize,LocalPixelmap);
            End;

            If Loaded
               Then Begin
                  AWidth:= tgWidth;
                  AHeight:= tgHeight;

                  CountPixel:= ImageSize Div 4;
                  RGBAPixel:= LocalPixelMap;
                  For I:= 0 To CountPixel-1 Do Begin
                     SwapRGBAPixel(RGBAPixel);
                     Inc (RGBAPixel);
                  End;

                  Result:= LocalPixelMap;
               End
               Else FreeMem(LocalPixelMap,ImageSize)
         End;

      LocalStream.Free;
   End;
End;
//********************************************************************
//********************************************************************
End.
