
UNIT dglRTTI;
// This unit contains:
// 1. ERTTIException = Exception for TRTTI
// 2. TRTTI_Handler = RTTI-handling object, can read/write every
//    object that was compiled with {$M+} or is a descendant of
//    TPersistent
//
// Not yet implemented:
//  - rounding of float numbers (concerns FPC)
//  - safety-functions for RTTI in writing methods

INTERFACE

USES
  SysUtils,
  Classes,
  TypInfo;

TYPE
  ERTTIException = Class(Exception);

  TRTTIValue = Record
    TypeKind: TTypeKind;

    VInteger: LongInt;    // tkInteger
    VFloat: Extended;     // tkFloat
    VOrdinal: Word;       // tkEnumeration
    VString: String;      // tkString,tkLString
    VObject: TObject;     // tkClass
    VMethod: TMethod;     // tkMethod
  End;

  TDataExchange = (dxBinary, dxString);

  TRTTI_Handler = Class(TObject)
  Private
    fClient       : TObject;
    fDataExchange : TDataExchange;
    fPropCount    : Integer;
    fPropList     : pPropList;

    Procedure SetClient(AClient: TObject);

    Function GetPropertyName(APropIndex: Integer): String;
    Function GetPropertyType(APropIndex: Integer): TTypeKind;

    Function GetValueByIndex(APropIndex: Integer): TRTTIValue;
    Function GetValueByName(APropName: String): TRTTIValue;

    Procedure SetValueByIndex(APropIndex: Integer;
                      ARTTIValue: TRTTIValue);
    Procedure SetValueByName(APropName: String;
                    ARTTIValue: TRTTIValue);

    Function GetStrValByIndex(APropIndex: Integer): String;
    Function GetStrValByName(APropName: String): String;

    Procedure SetStrValByIndex(APropIndex: Integer;
                      StrValue: String);
    Procedure SetStrValByName(APropName,StrValue: String);

    Function GetClassName: String;
    Function GetPropertyValue(APropInfo: PPropInfo): TRTTIValue;
    Procedure SetPropertyValue(APropInfo: PPropInfo;
                        AValue: TRTTIValue);
    Procedure OpenRTTI(AClient: TObject);
    Procedure CloseRTTI;
    Function RTTIAccessable: Boolean;
    Function ValidPropIndex(APropIndex: Integer): Boolean;
  Public
    Property Client: TObject Read fClient Write SetClient;
    Property DataExchange: TDataExchange
      Read fDataExchange Write fDataExchange;
    Property ClientClassName: String Read GetClassName;
    Property PropertyCount: Integer Read fPropCount;

    Property PropertyName[Index: Integer]: String
      Read GetPropertyName;
    Property PropertyType[Index: Integer]: TTypeKind
      Read GetPropertyType;

    Property ValueByIndex[Index: Integer]: TRTTIValue
      Read GetValueByIndex Write SetValueByIndex;
    Property ValueByName[Index: String]: TRTTIValue
      Read GetValueByName Write SetValueByName;

    Property StrValByIndex[Index: Integer]: String
      Read GetStrValByIndex Write SetStrValByIndex;
    Property StrValByName[Index: String]: String
      Read GetStrValByName Write SetStrValByName;

    Constructor Create; Reintroduce;
    Destructor Destroy; Override;
  End;


Procedure ClearRTTIValue(Var ARTTIValue: TRTTIValue);

IMPLEMENTATION

Constructor TRTTI_Handler.Create;
Begin
  Inherited Create;
  fClient:= Nil;
  fDataExchange:= dxBinary;
  fPropCount:= 0;
  fPropList:= Nil;
End;

Destructor TRTTI_Handler.Destroy;
Begin
  CloseRTTI;
  Inherited Destroy;
End;

Procedure TRTTI_Handler.SetClient(AClient: TObject);
Begin
  If Assigned(AClient) Then Begin // Client must not be Nil!
    OpenRTTI(AClient); // Provide RTTI-access of the new client
  End;
End;

Function TRTTI_Handler.GetClassName: String;
Begin
  Try
    If RTTIAccessable
      Then Result:= fClient.ClassName
      Else Result:= 'ERROR';
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetPropertyName(APropIndex: Integer): String;
Var
  PropInfo: PPropInfo;
Begin
  Try
    If RTTIAccessable And ValidPropIndex(APropIndex)
      Then Begin
        PropInfo:= fPropList[APropIndex];
        Result:= PropInfo.Name;
      End
      Else Result:= 'ERROR';
  Except
    On EAccessViolation Do CloseRTTI;
  End; 
End;

Function TRTTI_Handler.GetPropertyType(APropIndex: Integer): TTypeKind;
Var
  PropInfo: PPropInfo;
Begin
  Result:= tkUnknown;
  Try
    If RTTIAccessable And ValidPropIndex(APropIndex)
      Then Begin
        PropInfo:= fPropList[APropIndex];
        Result:= PropInfo.PropType^.Kind;
      End
      Else Result:= tkUnknown;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetValueByIndex(APropIndex: Integer): TRTTIValue;
Var
  PropInfo: PPropInfo;
Begin
  ClearRTTIValue(Result);
  Try
    If RTTIAccessable And ValidPropIndex(APropIndex)
      Then Begin
        PropInfo:= fPropList[APropIndex];
        Result:= GetPropertyValue(PropInfo);
      End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetValueByName(APropName: String): TRTTIValue;
Var
  PropInfo: PPropInfo;
Begin
  ClearRTTIValue(Result);
  Try
    If RTTIAccessable
      Then Begin
        PropInfo:= GetPropInfo(fClient, APropName);
        Result:= GetPropertyValue(PropInfo);
      End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Procedure TRTTI_Handler.SetValueByIndex(APropIndex: Integer;
                          ARTTIValue: TRTTIValue);
Var
  PropInfo: PPropInfo;
Begin
  Try
    If RTTIAccessable And ValidPropIndex(APropIndex)
      Then Begin
        PropInfo:= fPropList[APropIndex];
        SetPropertyValue(PropInfo, ARTTIValue);
      End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Procedure TRTTI_Handler.SetValueByName(APropName: String;
                        ARTTIValue: TRTTIValue);
Var
  PropInfo: PPropInfo;
Begin
  Try
    If RTTIAccessable
      Then Begin
        PropInfo:= GetPropInfo(fClient, APropName);
        SetPropertyValue(PropInfo, ARTTIValue);
      End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetStrValByIndex(APropIndex: Integer): String;
Begin
  Result:= '';
  Try
    If (fDataExchange = dxString) And (RTTIAccessable)
      Then Result:= GetValueByIndex(APropIndex).VString;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetStrValByName(APropName: String): String;
Begin
  Result:= '';
  Try
    If (fDataExchange = dxString) And (RTTIAccessable)
      Then Result:= GetValueByName(APropName).VString;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Procedure TRTTI_Handler.SetStrValByIndex(APropIndex: Integer;
                      StrValue: String);
Var LRTTIValue: TRTTIValue;
Begin
  Try
    If (fDataExchange = dxString) And (RTTIAccessable) Then Begin
      //ClearRTTIValue(LRTTIValue);
      //LRTTIValue.TypeKind:= tkString;
      LRTTIValue.VString:= StrValue;
      SetValueByIndex(APropIndex,LRTTIValue);
    End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Procedure TRTTI_Handler.SetStrValByName(APropName,StrValue: String);
Var LRTTIValue: TRTTIValue;
Begin
  Try
    If (fDataExchange = dxString) And (RTTIAccessable) Then Begin
      //ClearRTTIValue(LRTTIValue);
      //LRTTIValue.TypeKind:= tkString;
      LRTTIValue.VString:= StrValue;
      SetValueByName(APropName,LRTTIValue);
    End;
  Except
    On EAccessViolation Do CloseRTTI;
  End;
End;

Function TRTTI_Handler.GetPropertyValue(APropInfo: PPropInfo): TRTTIValue;
Begin
  ClearRTTIValue(Result);
  Result.TypeKind:= APropInfo.PropType^.Kind;

  With Result Do Case TypeKind Of

    tkInteger     : VInteger:= GetOrdProp(fClient,APropInfo);
    tkFloat       : VFloat  := GetFloatProp(fClient,APropInfo);
    tkEnumeration : VOrdinal:= GetOrdProp(fClient,APropInfo);
{$IFDEF FPC}
    tkSString,tkLString,tkAString:
              VString:= GetStrProp(fClient,APropInfo);
{$ELSE}
    tkString,tkLString:
              VString:= GetStrProp(fClient,APropInfo);
{$ENDIF}
    tkClass       : VObject:= GetObjectProp(fClient,APropInfo,TObject);
    tkMethod      : VMethod:= GetMethodProp(fClient,APropInfo);

    Else
      ERTTIException.Create('GetPropertyValue: TypeKind Error');

  End; // Of Case

  If fDataExchange = dxString Then
    With Result Do Case TypeKind Of

      tkInteger     : VString:= IntToStr(VInteger);
      tkFloat       : VString:= FloatToStr(VFloat);
      tkEnumeration : VString:= IntToStr(VOrdinal);
      tkClass       : VString:= '';
      tkMethod      : VString:= '';

    End; // Of Case
End;

Procedure TRTTI_Handler.SetPropertyValue(APropInfo: PPropInfo;
                          AValue: TRTTIValue);
Begin
  If fDataExchange = dxString Then
    With AValue Do Case APropInfo.PropType^.Kind Of

      tkInteger     : VInteger:= StrToInt(VString);
      tkFloat       : VFloat:= StrToFloat(VString);
      tkEnumeration : VOrdinal:= StrToInt(VString);
      tkClass       : VString:= '';
      tkMethod      : VString:= '';

    End; // Of Case

  With AValue Do Case APropInfo.PropType^.Kind Of

    tkInteger     : SetOrdProp(fClient,APropInfo,VInteger);
    tkFloat       : SetFloatProp(fClient,APropInfo,VFloat);
    tkEnumeration : SetOrdProp(fClient,APropInfo,VOrdinal);
{$IFDEF FPC}
    tkSString,tkLString,tkAString:
              SetStrProp(fClient,APropInfo,VString);
{$ELSE}
    tkString,tkLString:
              SetStrProp(fClient,APropInfo,VString);
{$ENDIF}
    tkClass       : SetObjectProp(fClient,APropInfo,VObject);
    tkMethod      : SetMethodProp(fClient,APropInfo,VMethod);

    Else
      ERTTIException.Create
        ('SetPropContent: TypeKind Error');

  End; // Of Case
End;

Procedure TRTTI_Handler.OpenRTTI(AClient: TObject);
Begin
  CloseRTTI;

  If AClient.ClassInfo <> Nil
    Then Begin
      fClient:= AClient;
      fPropCount:= GetTypeData(fClient.ClassInfo).PropCount;
      GetMem(fPropList, fPropCount * SizeOf(Pointer));
      GetPropList(fClient.ClassInfo,fPropList);
    End
    Else ERTTIException.Create
      ('OpenRTTI: No class information available');
End;

Procedure TRTTI_Handler.CloseRTTI;
Begin
  If Assigned(fPropList)
    Then Begin
      FreeMem(fPropList);
      fPropList:= Nil;
      fClient:= Nil;
    End;
End;

Function TRTTI_Handler.RTTIAccessable: Boolean;
Begin
  If (Assigned(fPropList)) And (fClient.ClassInfo <> Nil)
    Then Result:= True
    Else Result:= False;
End;

Function TRTTI_Handler.ValidPropIndex(APropIndex: Integer): Boolean;
Begin
  If (APropIndex >= 0) And (APropIndex < fPropCount)
    Then Result:= True
    Else Result:= False;
End;


Procedure ClearRTTIValue(Var ARTTIValue: TRTTIValue);
Begin
  With ARTTIValue Do Begin
    TypeKind:= tkUnknown;

    VInteger:= 0;
    VFloat  := 0;
    VOrdinal:= 0;
    VString := '';
    VObject := Nil;
    With VMethod Do Begin
      Code:= Nil;
      Data:= Nil;
    End;
  End;
End;


End.
