
UNIT dglDataProvider;
// This unit contains:
// 1. TDataProvider: partly abstract object for loading/saving the
//    published properties of TObjects.
//    Supported data format: see TRTTIValue (separate unit)
// 2. TCSV_Dataprovider, a concrete descendant of TDataProvider that
//    can load/save TObject-Data in CSV-format
//
// Not yet implemented:
//  - *** Errorhandling is not complete (try..except-blocks)
//  - TXML_Dataprovider
//  - TBinary_DataProvider
//  - The content of method pointers (events!)are not loaded/saved
//    as there is no information for finding them (more registration?)
//  - No "LoadChild"-function for loading trees yet
//  - downsize saved file by avoiding saving an object more than one
//    time (simply remember the object-pointer)

// Known problems:
// In Delphi: {$M+} is needed for the first ancestor only
// In FPC: every saveable/loadable object has to be compiled with {$M+}
// Ring joints between objects can cause endless loops when saving

INTERFACE

USES
  SysUtils,
  Classes,
  TypInfo,
  dglRTTI;

TYPE
  TFileAccess = (faLoad, faSave);

  TDGL_DataProvider = Class(TObject)
  Private
    fName: String;            // Filename without extension
    fFileAccess: TFileAccess; // Load/save
    fOpenSource: Boolean;     // Is DataSource accessable?
    fRTTI: TRTTI_Handler;        // RTTI-handling-object

    fClientClassname: String; // internal used variable

    Function CreateLoadedClass: TObject;
  Protected
    Function EndOfData: Boolean; Virtual; Abstract;
    Function GetFileName: String; Virtual; Abstract;
    Procedure CloseSource; Virtual; Abstract;
    Procedure OpenSource; Virtual; Abstract;
    Procedure LoadClientHeader; Virtual; Abstract;
    Procedure SaveClientHeader; Virtual; Abstract;
    Procedure LoadProperty(AIndex: Integer); Virtual; Abstract;
    Procedure SaveProperty(AIndex: Integer); Virtual; Abstract;
    Procedure SetDataFormat; Virtual; Abstract;
  Public
    Constructor Create(AName: String;
            AFileAccess: TFileAccess); Reintroduce;
    Destructor Destroy; Override;

    Function LoadClient: TObject;
    Procedure SaveClient(AClient: TObject);
  End;


  TCSV_DataProvider = Class(TDGL_DataProvider)
  Private
  Lines: TStringList;
  CurLine: Integer;
  Protected
  Function EndOfData: Boolean; Override;
  Function GetFileName: String; Override;
  Procedure CloseSource; Override;
  Procedure OpenSource; Override;
  Procedure LoadClientHeader; Override;
  Procedure SaveClientHeader; Override;
  Procedure LoadProperty(AIndex: Integer); Override;
  Procedure SaveProperty(AIndex: Integer); Override;
  Procedure SetDataFormat; Override;
  End;

Procedure CrackUpString(AString: String; AStringList: TStringList;
      ASeparator: Char; Const ConsiderEmptyStrings: Boolean);

IMPLEMENTATION

// CrackupString cracks a C(omma) S(eparated) V(alue) - string to its
// several parts and fills them in a stringlist
Procedure CrackUpString(AString: String; AStringList: TStringList;
      ASeparator: Char; Const ConsiderEmptyStrings: Boolean);
Var
  FirstPos, LinePos, ItemLen, StrgLen: Word;
  Interim: String;
//--------------------------------------------
Function EOS: Boolean; // End Of String
Begin
  If LinePos <= StrgLen
    Then Result:= False
    Else Result:= True;
End;
//--------------------------------------------
Function EOI: Boolean; // End Of Single Item
Begin
  Result:= True;
  If Not EOS Then
  If AString[LinePos] <> ASeparator
    Then Result:= False
    Else Result:= True;
End;
//--------------------------------------------
Begin
  AStringList.Clear; StrgLen:= Length(AString); LinePos:= 1;
  While Not EOS Do Begin
    FirstPos:= LinePos;
    // scroll to Item-End
    While (Not EOS) And (Not EOI) Do Inc(LinePos);
    // set string, copy and put in list
    ItemLen:= LinePos - FirstPos;
    If (ItemLen > 0) Or (ConsiderEmptyStrings) Then
    Begin
      SetLength(Interim,ItemLen);
      If ItemLen > 0 Then
        Move(AString[FirstPos],Interim[1],ItemLen);
      AStringList.Add(Interim);
    End;
    Inc(LinePos); // = Separator berspringen
  End;
End;

// TDGL_DataProvider

Constructor TDGL_DataProvider.Create(AName: String;
                    AFileAccess: TFileAccess);
Begin
  Inherited Create;
  fName:= AName;
  fFileAccess:= AFileAccess;
  fRTTI:= TRTTI_Handler.Create;
  fOpenSource:= False;
  OpenSource;
End;

Destructor TDGL_DataProvider.Destroy;
Begin
  CloseSource;
  fRTTI.Free;
  Inherited Destroy;
End;

Function TDGL_DataProvider.CreateLoadedClass: TObject;
Var
  LoadedClass: TClass;
  LocalClient: TObject;
  I: Integer;
Begin
  Result:= Nil;
  LoadedClass:= FindClass(fClientClassName);
  If LoadedClass <> Nil Then
  Begin
    LocalClient:= LoadedClass.Create;
    SetDataFormat; // String or binary

    fRTTI.Client:= LocalClient;
    With fRTTI Do
      If PropertyCount > 0 Then
      For I:= 0 To PropertyCount-1 Do
        LoadProperty(I);

    Result:= LocalClient;
  End;
End;

Function TDGL_DataProvider.LoadClient: TObject;
Begin
  Result:= Nil;
  If fOpenSource Then
  Begin
    LoadClientHeader; // = Load ClassName
    Result:= CreateLoadedClass;
  End;
End;

Procedure TDGL_DataProvider.SaveClient(AClient: TObject);
var
I: Integer;
Begin
  fRTTI.Client:= AClient;
  SaveClientHeader; // = Save ClassName
  SetDataFormat;    // String or Binary

  With fRTTI Do If PropertyCount > 0 Then
  For I:= 0 To PropertyCount-1 Do
    SaveProperty(I);
End;


//   TCSV_DataProvider

Function TCSV_DataProvider.GetFileName: String;
Begin
  Result:= fName + '.CSV';
End;

Procedure TCSV_DataProvider.OpenSource;
Begin
  Lines:= TStringList.Create;

  If fFileAccess = faLoad Then
    If FileExists(GetFileName) Then
    Begin
      Lines.LoadFromFile(GetFileName);
      fOpenSource:= True;
    End;

  CurLine:= 0;
End;

Procedure TCSV_DataProvider.CloseSource;
Begin
  If fFileAccess = faSave Then
    Lines.SaveToFile(GetFileName);

  Lines.Free;
End;

Function TCSV_DataProvider.EndOfData: Boolean;
Begin
  If (CurLine >= Lines.Count)
    Then Result:= TRUE
    Else Result:= FALSE;
End;

Procedure TCSV_DataProvider.LoadClientHeader;
Begin
  fClientClassName:= Lines[CurLine];
  Delete(fClientClassName,Length(fClientClassName),1);
  Inc(CurLine);
End;

Procedure TCSV_DataProvider.SaveClientHeader;
Begin
  Lines.Add(fRTTI.ClientClassName + ';');
End;

Procedure TCSV_DataProvider.LoadProperty(AIndex: Integer);
Var
  OldClient: TObject;
  CrackList: TStringList;
  LocalRTTIValue: TRTTIValue;
Begin
  Case fRTTI.PropertyType[AIndex] Of
  //-------------------------------------------------------
  tkClass:
    Begin
      ClearRTTIValue(LocalRTTIValue);
      LocalRTTIValue.TypeKind:= tkClass;

      OldClient:= fRTTI.Client;            // Save Client
      LocalRTTIValue.VObject:= LoadClient; // Load SubObject from source
      fRTTI.Client:= OldClient;            // Restore Client
      fRTTI.ValueByIndex[AIndex]:= LocalRTTIValue; // Add SubObject
    End;                                            //    to Client
  //-------------------------------------------------------
  tkMethod: ;
  //-------------------------------------------------------
  Else
    Begin
      CrackList:= TStringList.Create;
      CrackUpString(Lines[CurLine],CrackList,';',True);

      fRTTI.StrValByIndex[AIndex]:= CrackList[1];
      Inc(CurLine);

      CrackList.Free;
    End;
  End; // Of Case;
End;

Procedure TCSV_DataProvider.SaveProperty(AIndex: Integer);
Var
  OldClient: TObject;
  LocalRTTIValue: TRTTIValue;
Begin
  Case fRTTI.PropertyType[AIndex] Of
  //-------------------------------------------------------
  tkClass:
    Begin
      OldClient:= fRTTI.Client;             // Save Client
      LocalRTTIValue:= fRTTI.ValueByIndex[AIndex];  // Get PropValue
      SaveClient(LocalRTTIValue.VObject);   // Save Client to source
      fRTTI.Client:= OldClient;             // Restore Client
    End;
  //-------------------------------------------------------
  tkMethod: ;
  //-------------------------------------------------------
  Else
    With fRTTI Do Begin
      Lines.Add(PropertyName[AIndex] + ';'
          + StrValByIndex[AIndex] + ';');
    End;
  End; // Of Case;
End;

Procedure TCSV_DataProvider.SetDataFormat;
Begin
  fRTTI.DataExchange:= dxString;
End;

End.
